import asyncio
import ssl

import anyio
import httpx
from pypika import PostgreSQLQuery
from pypika import functions as pypika_fn

from ffun.core import logging
from ffun.core.postgresql import execute
from ffun.domain.entities import AbsoluteUrl, FeedUrl
from ffun.feeds.entities import FeedError
from ffun.loader import errors
from ffun.loader.entities import ProxyState
from ffun.loader.settings import Proxy, settings
from ffun.parsers import entities as p_entities
from ffun.parsers.domain import parse_feed

logger = logging.get_module_logger()


_load_semaphor = asyncio.Semaphore(settings.max_concurrent_http_requests)

_accept_enconding_header = "br;q=1.0, zstd;q=0.9, gzip;q=0.8, deflate;q=0.7"


async def load_content(  # noqa: CFQ001, CCR001, C901 # pylint: disable=R0912, R0915
    url: AbsoluteUrl, proxy: Proxy, user_agent: str, semaphore: asyncio.Semaphore = _load_semaphor
) -> httpx.Response:
    error_code = FeedError.network_unknown

    log = logger.bind(url=url, proxy=proxy.name, function="load_content", user_agent=user_agent)

    try:
        log.info("loading_feed")

        headers = {"user-agent": user_agent, "accept-encoding": _accept_enconding_header}

        async with semaphore:
            async with httpx.AsyncClient(proxy=proxy.url, headers=headers) as client:
                response = await client.get(url, follow_redirects=True)

    except httpx.RemoteProtocolError as e:
        message = str(e)

        if "illegal request line" in message:
            # TODO: at least part of such errors are caused by wrong HTTP protocol
            #       for example `http://gopractice.ru/feed/` tries to redirect to use HTTP/0.9
            log.warning("network_illegal_request_line")
            error_code = FeedError.network_illegal_request_line
        elif "Server disconnected without sending a response" in message:
            log.warning("network_disconection_without_response")
            error_code = FeedError.network_disconection_without_response
        elif "peer closed connection without sending complete message body" in message:
            log.warning("network_received_unkomplete_body")
            error_code = FeedError.network_received_unkomplete_body
        elif "Invalid URL in location header" in message:
            # Handle some incorrect redirects, for example, caused by the Great China Firewall
            # in the case of GCF you can see the "fzyujing" in new location, but it does not returned by httpx
            # => we just check the base message
            # Details: https://chatgpt.com/share/6855377e-4828-800d-a4cb-1ea924cd6286
            error_code = FeedError.network_wrong_redirect
        else:
            log.exception("remote_protocol_error_while_loading_feed")

        raise errors.LoadError(feed_error_code=error_code) from e

    except httpx.ReadError as e:
        message = str(e)

        if message == "":
            error_code = FeedError.network_read_error
            log.warning("network_read_error")
        else:
            log.exception("unknown_read_error_while_loading_feed")

        raise errors.LoadError(feed_error_code=error_code) from e

    except httpx.ConnectError as e:
        message = str(e)

        if "[Errno -2]" in message:
            log.warning("network_name_or_service_not_known")
            error_code = FeedError.network_name_or_service_not_known
        elif "[Errno -3]" in message:
            log.warning("network_temporary_failure_in_name_resolution")
            error_code = FeedError.network_temporary_failure_in_name_resolution
        elif "[Errno -5]" in message:
            log.warning("no_address_associated_with_hostname")
            error_code = FeedError.network_no_address_associated_with_hostname
        elif message == "":
            log.warning("undetected_connection_error")
            error_code = FeedError.network_undetected_connection_error
        elif message == "All connection attempts failed":
            log.warning("network_all_connection_attempts_failed")
            error_code = FeedError.network_all_connection_attempts_failed
        elif "[SSL:" in message:
            # catch all SSL errors
            log.warning("network_ssl_connection_error")
            error_code = FeedError.network_ssl_connection_error
        else:
            log.exception("connection_error_while_loading_feed")

        raise errors.LoadError(feed_error_code=error_code) from e

    except ssl.SSLError as e:
        # TODO: it is possible, that httpx started to wrap ssl errors into httpx.ConnectError
        #       for example, see CERTIFICATE_VERIFY_FAILED.
        #       If it is true, we may remove this block
        log.warning("network_certificate_verify_failed")
        error_code = FeedError.network_ssl_connection_error
        raise errors.LoadError(feed_error_code=error_code) from e

    except ssl.SSLCertVerificationError as e:
        message = str(e)

        # TODO: it is possible, that httpx started to wrap ssl errors into httpx.ConnectError
        #       for example, see CERTIFICATE_VERIFY_FAILED.
        #       If it is true, we may remove this block
        if "CERTIFICATE_VERIFY_FAILED" in message:
            log.warning("network_certificate_verify_failed")
            error_code = FeedError.network_certificate_verify_failed
        else:
            log.exception("ssl_cert_verification_error_while_loading_feed")

        raise errors.LoadError(feed_error_code=error_code) from e

    except httpx.ConnectTimeout as e:
        log.warning("network_connect_timeout")
        error_code = FeedError.network_connection_timeout
        raise errors.LoadError(feed_error_code=error_code) from e

    except httpx.ReadTimeout as e:
        log.warning("network_read_timeout")
        error_code = FeedError.network_read_timeout
        raise errors.LoadError(feed_error_code=error_code) from e

    except httpx.UnsupportedProtocol as e:
        log.warning("network_unsupported_protocol")
        error_code = FeedError.network_unsupported_protocol
        raise errors.LoadError(feed_error_code=error_code) from e

    except httpx.TooManyRedirects as e:
        log.warning("network_too_many_redirects")
        error_code = FeedError.network_too_many_redirects
        raise errors.LoadError(feed_error_code=error_code) from e

    except anyio.EndOfStream as e:
        log.warning("server_breaks_connection")
        error_code = FeedError.network_server_breaks_connection
        raise errors.LoadError(feed_error_code=error_code) from e

    except httpx.ProxyError as e:
        message = str(e)

        if message.startswith("502 Could not resolve host"):
            log.warning("proxy_could_not_resolve_host")
            error_code = FeedError.proxy_could_not_resolve_host
        elif "TUN_ERR" in message and "ECONNREFUSED" in message:
            log.warning("proxy_connection_refused")
            error_code = FeedError.proxy_connection_refused
        elif "TUN_ERR" in message and "EHOSTUNREACH" in message:
            log.warning("proxy_no_route_to_host")
            error_code = FeedError.proxy_no_route_to_host
        elif "403" in message:
            log.warning("proxy_connection_403")
            error_code = FeedError.proxy_connection_403
        elif "502" in message:
            log.warning("proxy_connection_502")
            error_code = FeedError.proxy_connection_502
        elif "429" in message:
            log.warning("proxy_connection_429")
            error_code = FeedError.proxy_connection_429
        else:
            log.exception("unknown_proxy_error_while_loading_feed")

        raise errors.LoadError(feed_error_code=error_code) from e

    except httpx.DecodingError as e:
        log.warning("network_decoding_error")
        error_code = FeedError.network_decoding_error
        raise errors.LoadError(feed_error_code=error_code) from e

    except Exception as e:
        log.exception("error_while_loading_feed")
        raise errors.LoadError(feed_error_code=error_code) from e

    if response.status_code != 200:
        log.warning("network_non_200_status_code", status_code=response.status_code)
        error_code = FeedError.network_non_200_status_code
        raise errors.LoadError(feed_error_code=error_code)

    log.info("feed_loaded", url=url, proxy=proxy.name)

    return response


# TODO: tests
async def decode_content(response: httpx.Response) -> str:
    error_code = FeedError.parsing_base_error

    encoding = response.encoding or "utf-8"

    try:
        return response.content.decode(encoding)
    except UnicodeDecodeError as e:
        logger.warning("unicode_decode_error_while_decoding_feed")
        error_code = FeedError.parsing_unicode_decode_error
        raise errors.LoadError(feed_error_code=error_code) from e
    except Exception as e:
        logger.exception("error_while_decoding_feed")
        raise errors.LoadError(feed_error_code=error_code) from e


# TODO: tests
async def parse_content(content: str, original_url: FeedUrl) -> p_entities.FeedInfo:
    try:
        feed_info = parse_feed(content, original_url=original_url)
    except Exception as e:
        logger.exception("error_while_parsing_feed")
        raise errors.LoadError(feed_error_code=FeedError.parsing_format_error) from e

    if feed_info is None:
        raise errors.LoadError(feed_error_code=FeedError.parsing_feed_content_not_found)

    if not feed_info.entries:
        raise errors.LoadError(feed_error_code=FeedError.protocol_no_entries_in_feed)

    return feed_info


async def check_proxy(proxy: Proxy, url: str, user_agent: str) -> bool:
    try:
        async with httpx.AsyncClient(proxy=proxy.url, headers={"user-agent": user_agent}) as client:
            response = await client.head(url)
    except Exception as e:
        logger.info("proxy_check_error", proxy=proxy.name, url=url, error=str(e))
        return False

    logger.info("proxy_check", proxy=proxy.name, url=url, status_code=response.status_code)

    return response.status_code == 200


async def is_proxy_available(proxy: Proxy, anchors: list[str], user_agent: str) -> bool:
    tasks = [check_proxy(proxy, url, user_agent) for url in anchors]

    results = await asyncio.gather(*tasks, return_exceptions=True)

    return any(result for result in results)


async def get_proxy_states(names: list[str]) -> dict[str, ProxyState]:
    sql = """
    SELECT name, state
    FROM lr_proxy_states
    WHERE name = ANY(%(names)s)
    """

    rows = await execute(sql, {"names": names})

    states = {name: ProxyState.available for name in names}

    for row in rows:
        states[row["name"]] = ProxyState(row["state"])

    return states


async def update_proxy_states(states: dict[str, ProxyState]) -> None:
    if not states:
        return

    query = PostgreSQLQuery.into("lr_proxy_states").columns("name", "state", "updated_at")

    for name, state in states.items():
        query = query.insert(name, state, pypika_fn.Now())

    query = query.on_conflict("name").do_update("state").do_update("updated_at", pypika_fn.Now())

    await execute(str(query))



"""
processed-state
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__ = {"20230427_01_pv33u-fix-entries-unique-index"}


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("ALTER TABLE l_entry_process_info ADD COLUMN state SMALLINT NOT NULL DEFAULT 1")
    cursor.execute("ALTER TABLE l_entry_process_info ADD last_error TEXT")


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("ALTER TABLE l_entry_process_info DROP COLUMN state")
    cursor.execute("ALTER TABLE l_entry_process_info DROP COLUMN last_error")


steps = [step(apply_step, rollback_step)]



"""
replace tokens traking with costs tracking
"""

import datetime
from typing import Any

from psycopg import Connection
from psycopg.rows import dict_row
from yoyo import step

__depends__ = {"20230702_01_LEEES-resources-table"}


def month_interval_start() -> datetime.datetime:
    now = datetime.datetime.now(tz=datetime.timezone.utc)

    return now.replace(day=1, hour=0, minute=0, second=0, microsecond=0)


sql = """
INSERT INTO r_resources (user_id, kind, interval_started_at, used, reserved)
VALUES (%(user_id)s, 2, %(interval_started_at)s, %(used)s, %(reserved)s)
"""


k = 1_000_000_000
cost_1m = 0.6


def tokens_to_points(tokens: int) -> int:
    return int((tokens / 1_000_000 * cost_1m) * k)


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor(row_factory=dict_row)

    interval = month_interval_start()

    cursor.execute(
        "SELECT * FROM r_resources WHERE kind=1 AND interval_started_at=%(interval)s", {"interval": interval}
    )

    for row in cursor.fetchall():
        cursor.execute(
            sql,
            {
                "user_id": row["user_id"],
                "interval_started_at": row["interval_started_at"],
                "used": tokens_to_points(row["used"]),
                "reserved": tokens_to_points(row["reserved"]),
            },
        )

    cursor.execute("DELETE FROM r_resources WHERE kind=1")


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    pass


steps = [step(apply_step, rollback_step)]



import uuid

from ffun.domain.entities import CollectionId, EntryId, FeedId, RuleId, SourceId, UserId


def no_user_id() -> UserId:
    return UserId(uuid.UUID(int=0))


def new_user_id() -> UserId:
    return UserId(uuid.uuid4())


def new_entry_id() -> EntryId:
    return EntryId(uuid.uuid4())


def new_feed_id() -> FeedId:
    return FeedId(uuid.uuid4())


def new_collection_id() -> CollectionId:
    return CollectionId(uuid.uuid4())


def new_source_id() -> SourceId:
    return SourceId(uuid.uuid4())


def new_rule_id() -> RuleId:
    return RuleId(uuid.uuid4())



"""
turn-on-unique-uids
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__ = {"20240504_02_gEapd-fill-uids-for-feeds"}


unique_index = "idx_f_feeds_uid"


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()

    cursor.execute("ALTER TABLE f_feeds ALTER COLUMN uid SET NOT NULL")
    cursor.execute(f"ALTER TABLE f_feeds ADD CONSTRAINT {unique_index} UNIQUE (uid)")


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute(f"ALTER TABLE f_feeds DROP CONSTRAINT {unique_index}")
    cursor.execute("ALTER TABLE f_feeds ALTER COLUMN uid DROP NOT NULL")


steps = [step(apply_step, rollback_step)]



import re

from bs4 import BeautifulSoup, Tag


def clear_nothing(text: str) -> str:
    return text


def clear_html(text: str) -> str:
    soup = BeautifulSoup(text, "html.parser")

    for tag in soup(["script", "style", "meta", "iframe", "img"]):
        tag.decompose()

    for tag in soup():

        if not isinstance(tag, Tag):
            continue

        if tag.name in {"h1", "h2", "h3", "h4", "h5", "h6", "a", "p", "li", "ul", "ol"}:
            continue

        tag.unwrap()

    simplified_html = str(soup)

    # remove borken surrogate unicode characters
    unicoded_html = simplified_html.encode("utf-16", "surrogatepass").decode("utf-16", "ignore")

    cleaned_text = re.sub(r"\s+", " ", unicoded_html)

    return cleaned_text.strip()



import pathlib

import pydantic_settings

from ffun.core.settings import BaseSettings

_root = pathlib.Path(__file__).parent


class Settings(BaseSettings):

    api_entry_point: str | None = None
    api_timeout: float = 60.0

    model_config = pydantic_settings.SettingsConfigDict(env_prefix="FFUN_OPENAI_")


settings = Settings()



from typing import Annotated

import fastapi
from supertokens_python.recipe.session import SessionContainer
from supertokens_python.recipe.session.framework.fastapi import verify_session

from ffun.auth.settings import AuthMode, settings
from ffun.users import domain as u_domain
from ffun.users import entities as u_entities

# ATTENTION: We check the database in supertokens sessions to be able to logout user from backend from all sessions.
#            It is the fastest and easiest way to do it, and it should not cause performance issues, at least for now.
#            In the future, we may want to improve user experience by removing this check and
#            implementing faster custom logout


async def _supertokens_user(
    session: SessionContainer = fastapi.Depends(verify_session(check_database=True)),
) -> u_entities.User:
    return await u_domain.get_or_create_user(u_entities.Service.supertokens, session.user_id)


async def _supertokens_optional_user(
    session: SessionContainer = fastapi.Depends(verify_session(check_database=True, session_required=False)),
) -> u_entities.User | None:
    if session is None:
        return None

    return await u_domain.get_or_create_user(u_entities.Service.supertokens, session.user_id)


async def _single_user() -> u_entities.User:
    return await u_domain.get_or_create_user(u_entities.Service.single, settings.single_user.external_id)


if settings.mode == AuthMode.single_user:
    user = fastapi.Depends(_single_user)
    optional_user = fastapi.Depends(_single_user)
elif settings.mode == AuthMode.supertokens:
    user = fastapi.Depends(_supertokens_user)
    optional_user = fastapi.Depends(_supertokens_optional_user)
else:
    raise NotImplementedError(f"AuthMode {settings.mode} not implemented")


User = Annotated[u_entities.User, user]

OptionalUser = Annotated[u_entities.User | None, optional_user]



import datetime

from ffun.domain import urls as d_urls
from ffun.domain.entities import EntryId, FeedId, UnknownUrl
from ffun.feeds import domain as f_domain
from ffun.library import operations
from ffun.library.entities import Entry, EntryChange

catalog_entries = operations.catalog_entries
get_entries_by_ids = operations.get_entries_by_ids
get_entries_by_filter = operations.get_entries_by_filter
find_stored_entries_for_feed = operations.find_stored_entries_for_feed
all_entries_iterator = operations.all_entries_iterator
get_entries_after_pointer = operations.get_entries_after_pointer
unlink_feed_tail = operations.unlink_feed_tail
get_feed_links_for_entries = operations.get_feed_links_for_entries
get_orphaned_entries = operations.get_orphaned_entries
remove_entries_by_ids = operations.remove_entries_by_ids
count_total_entries = operations.count_total_entries


async def get_entry(entry_id: EntryId) -> Entry:
    entries = await get_entries_by_ids([entry_id])
    found_entry = entries.get(entry_id)
    assert found_entry is not None
    return found_entry


async def get_feeds_for_entry(entry_id: EntryId) -> set[FeedId]:
    mapping = await get_feed_links_for_entries([entry_id])

    if not mapping:
        return set()

    return {link.feed_id for link in mapping[entry_id]}


async def normalize_entry(entry: Entry, apply: bool = False) -> list[EntryChange]:
    feed_links_mapping = await get_feed_links_for_entries([entry.id])

    if entry.id not in feed_links_mapping:
        return []

    feed_links = feed_links_mapping[entry.id]

    feed_links.sort(key=lambda link: link.created_at)

    # use oldest link to chose normalization feed
    # TODO: give priority to feeds from collections
    feed_id = feed_links[0].feed_id

    feed = await f_domain.get_feed(feed_id)

    new_external_url = d_urls.adjust_external_url(UnknownUrl(entry.external_url), feed.url)

    changes = []

    if new_external_url != entry.external_url:
        changes.append(
            EntryChange(id=entry.id, field="external_url", old_value=entry.external_url, new_value=new_external_url)
        )
        if apply and new_external_url is not None:
            await operations.update_external_url(entry.id, new_external_url)

    return changes


async def get_entries_by_filter_with_fallback(
    feeds_ids: list[FeedId], period: datetime.timedelta | None, limit: int, fallback_limit: int
) -> list[Entry]:

    entries = await get_entries_by_filter(feeds_ids=feeds_ids, period=period, limit=limit)

    if entries:
        return entries

    # if there is no news in requested interval try to get some older news
    entries = await get_entries_by_filter(
        feeds_ids=feeds_ids, period=datetime.timedelta(days=10 * 365), limit=fallback_limit
    )

    return entries



"""
remove-duplicated-entries-from-feeds
"""

from typing import Any

from psycopg import Connection
from psycopg.rows import dict_row
from yoyo import step

__depends__ = {"20240503_01_Bmzw6-remove-l-entry-process-info"}


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor(row_factory=dict_row)

    sql = "SELECT feed_id, external_id FROM l_entries GROUP BY feed_id, external_id HAVING count(*) > 1"

    cursor.execute(sql)

    ids_to_remove = []

    for row in cursor.fetchall():
        feed_id = row["feed_id"]
        external_id = row["external_id"]

        cursor.execute(
            """
            SELECT id FROM l_entries
            WHERE feed_id = %(feed_id)s AND external_id = %(external_id)s
            ORDER BY created_at ASC
            """,
            {"feed_id": feed_id, "external_id": external_id},
        )

        result = cursor.fetchall()

        ids_to_remove.extend([row["id"] for row in result[1:]])

    sql = "DELETE FROM l_entries WHERE id = ANY(%(ids)s)"

    cursor.execute(sql, {"ids": ids_to_remove})


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    pass


steps = [step(apply_step, rollback_step)]



import pydantic
import pytest

from ffun.librarian.entities import LLMGeneralProcessor
from ffun.librarian.tag_extractors import dog_tags_extractor
from ffun.librarian.text_cleaners import clear_nothing
from ffun.llms_framework.entities import LLMApiKey, LLMConfiguration, LLMGeneralApiKey, LLMProvider, LLMTokens


class TestLLMGeneralProcessor:

    @pytest.fixture
    def llm_config(self) -> LLMConfiguration:
        return LLMConfiguration(
            model="test-model",
            system="some system prompt",
            max_return_tokens=LLMTokens(1017),
            text_parts_intersection=113,
            temperature=0.3,
            top_p=0.9,
            presence_penalty=0.5,
            frequency_penalty=0.75,
        )

    def test_api_key_is_required_if_collections_enbabled(self, llm_config: LLMConfiguration) -> None:
        with pytest.raises(pydantic.ValidationError) as exc_info:
            LLMGeneralProcessor(
                id=666,
                enabled=True,
                workers=1,
                name="test-processor",
                allowed_for_collections=True,
                allowed_for_users=False,
                llm_provider=LLMProvider.test,
                llm_config=llm_config,
                entry_template="<h1>{entry.title}</h1><a href='{entry.external_url}'>full article</a>{entry.body}",
                text_cleaner=clear_nothing,
                tags_extractor=dog_tags_extractor,
                collections_api_key=None,
                general_api_key=None,
            )

        assert exc_info.value.errors()[0]["type"] == "collections_or_general_key_required_for_collections"

    @pytest.mark.parametrize("key_warning", [None, "wrong warning"])
    def test_general_api_key_warning_check__failed(
        self, llm_config: LLMConfiguration, key_warning: str | None
    ) -> None:
        with pytest.raises(pydantic.ValidationError) as exc_info:
            LLMGeneralProcessor(
                id=666,
                enabled=True,
                workers=1,
                name="test-processor",
                allowed_for_collections=False,
                allowed_for_users=False,
                llm_provider=LLMProvider.test,
                llm_config=llm_config,
                entry_template="<h1>{entry.title}</h1><a href='{entry.external_url}'>full article</a>{entry.body}",
                text_cleaner=clear_nothing,
                tags_extractor=dog_tags_extractor,
                collections_api_key=None,
                general_api_key=LLMGeneralApiKey(LLMApiKey("some key")),
                general_api_key_warning=key_warning,
            )

        assert exc_info.value.errors()[0]["type"] == "you_must_confirm_general_api_key_usage"



import pytest
from furl import furl

from ffun.domain import errors, urls
from ffun.domain.entities import AbsoluteUrl, FeedUrl, SourceUid, UnknownUrl, UrlUid


class TestAdjustClassicUrl:
    @pytest.mark.parametrize(
        "original_url, raw_url, adjusted_url",
        [
            ("https://example.com/feed/", "https://example.com", "https://example.com"),
            ("https://example.com/feed/", "https://example.com/path/a/b?c=d", "https://example.com/path/a/b?c=d"),
            (
                "https://example.com/feed/",
                "http://another.domain.com:666/path/a/b?c=d",
                "http://another.domain.com:666/path/a/b?c=d",
            ),
            (
                "https://example.com/feed/",
                "another.domain.com:666/path/a/b?c=d",
                "https://another.domain.com:666/path/a/b?c=d",
            ),
            (
                "https://example.com/feed/",
                "another.domain.com/path/a/b?c=d",
                "https://another.domain.com/path/a/b?c=d",
            ),
            ("https://example.com/feed/", "/path/a/b?c=d", "https://example.com/path/a/b?c=d"),
            ("https://example.com/feed/", "path/a/b?c=d", "https://example.com/feed/path/a/b?c=d"),
            ("https://example.com/feed/", "path", "https://example.com/feed/path"),
            ("https://example.com/feed/", "?c=d", "https://example.com/feed/?c=d"),
            ("https://example.com/feed/", "another.domain.com", "https://another.domain.com"),
            ("https://example.com", "another.domain.com/path/a/b?c=d", "https://another.domain.com/path/a/b?c=d"),
            ("https://example.com", "/path/a/b?c=d", "https://example.com/path/a/b?c=d"),
            ("https://example.com", "path/a/b?c=d", "https://example.com/path/a/b?c=d"),
            ("https://example.com", "path", "https://example.com/path"),
            ("https://example.com", "?c=d", "https://example.com?c=d"),
            ("https://example.com", "another.domain.com", "https://another.domain.com"),
            ("//example.com/feed/", "https://example.com", "https://example.com"),
            ("//example.com/feed/", "example.com/path/a/b?c=d", "//example.com/path/a/b?c=d"),
            (
                "//example.com/feed/",
                "http://another.domain.com:666/path/a/b?c=d",
                "http://another.domain.com:666/path/a/b?c=d",
            ),
            ("//example.com/feed/", "another.domain.com:666/path/a/b?c=d", "//another.domain.com:666/path/a/b?c=d"),
            ("//example.com/feed/", "another.domain.com/path/a/b?c=d", "//another.domain.com/path/a/b?c=d"),
            ("//example.com/feed/", "/path/a/b?c=d", "//example.com/path/a/b?c=d"),
            ("//example.com/feed/", "path/a/b?c=d", "//example.com/feed/path/a/b?c=d"),
            ("//example.com/feed/", "path", "//example.com/feed/path"),
            ("//example.com/feed/", "?c=d", "//example.com/feed/?c=d"),
            ("//example.com/feed/", "another.domain.com", "//another.domain.com"),
            # test keeping fragment
            (
                "https://example.com/feed/",
                "https://example.com/path/a/b?c=d#fragment",
                "https://example.com/path/a/b?c=d#fragment",
            ),
            ("//example.com/feed/", "/path/a/b?c=d#z=q", "//example.com/path/a/b?c=d#z=q"),
        ],
    )
    def test_base_transformations(
        self, original_url: AbsoluteUrl, raw_url: UnknownUrl, adjusted_url: AbsoluteUrl
    ) -> None:
        assert urls.is_absolute_url(original_url)
        assert urls.is_absolute_url(adjusted_url)

        assert urls.adjust_classic_url(raw_url, original_url) == adjusted_url


class TestIsMagneticUrl:
    @pytest.mark.parametrize(
        "url, is_magnetic",
        [
            (
                "magnet:?xt=urn:btih:123456789abcdef0123456789abcdef0123456789&dn=Example+File.mp4&tr=udp%3A%2F%2Ftracker.example.com%3A80",  # noqa
                True,
            ),
            ("https://example.com/path/a/b?c=d", False),
            ("http://another.domain.com:666/path/a/b?c=d", False),
        ],
    )
    def test(self, url: UnknownUrl, is_magnetic: bool) -> None:
        assert urls.is_magnetic_url(url) == is_magnetic


class TestAdjustMagneticUrl:
    def test(self) -> None:
        url = "magnet:?xt=urn:btih:123456789abcdef0123456789abcdef0123456789&dn=Example+File.mp4&tr=udp%3A%2F%2Ftracker.example.com%3A80"  # noqa
        assert urls.adjust_magnetic_url(UnknownUrl(url)) == url


class TestAdjustExternalUrl:
    @pytest.mark.parametrize(
        "url, adjusted_url",
        [
            ("https://example.com/path/a/b?c=d", "https://example.com/path/a/b?c=d"),
            ("http://another.domain.com:666/path/a/b?c=d", "http://another.domain.com:666/path/a/b?c=d"),
            ("/path/a/b?c=d", "https://example.com/path/a/b?c=d"),
            (
                "magnet:?xt=urn:btih:123456789abcdef0123456789abcdef0123456789&dn=Example+File.mp4&tr=udp%3A%2F%2Ftracker.example.com%3A80",  # noqa
                "magnet:?xt=urn:btih:123456789abcdef0123456789abcdef0123456789&dn=Example+File.mp4&tr=udp%3A%2F%2Ftracker.example.com%3A80",  # noqa
            ),
            ("http://www.usinenouvelle.comhttps://www.usine-digitale.fr/article/christophe", None),
            ("x" * 100, "https://example.com/" + "x" * 100),  # noqa
            ("x-x" * 100, "https://example.com/" + "x-x" * 100),  # noqa
            ("xxx.com", "https://xxx.com"),  # noqa
            ("x" * 100 + ".html", "https://example.com/" + "x" * 100 + ".html"),  # noqa
            ("x.html", "https://example.com/x.html"),  # noqa
            # TODO: what if ip addreses?
        ],
    )
    def test(self, url: UnknownUrl, adjusted_url: AbsoluteUrl) -> None:
        assert (
            adjusted_url is None or urls.is_absolute_url(adjusted_url) or urls.is_magnetic_url(url)
        ), "Wrong test parameters"

        original_url = urls.normalize_classic_unknown_url(UnknownUrl("https://example.com"))

        assert original_url is not None

        assert urls.adjust_external_url(url, original_url=original_url) == adjusted_url


class TestUrlToUid:
    # This parameters are a bit excessive
    # Current implementation of url_to_uid expects already normalized AbsoluteUrl
    @pytest.mark.parametrize(
        "url, uid",
        [  # strip spaces
            ("   //example.com", "example.com"),
            ("//example.com   ", "example.com"),
            # trailing slash
            ("//example.com/", "example.com"),
            ("//example.com//", "example.com"),
            ("//example.com/path/", "example.com/path"),
            # uppercase
            ("//EXAMPLE.COM", "example.com"),
            ("//exaMple.com/somE/pAth", "example.com/some/path"),
            # pluses
            ("//example.com/with+plus", "example.com/with+plus"),
            ("//example.com/with%20space", "example.com/with+space"),
            ("//example.com/with plus", "example.com/with+plus"),
            # quotes
            ("//example.com/with%22quote", 'example.com/with"quote'),
            ("//example.com/with%27quote", "example.com/with'quote"),
            # query
            ("//example.com/?a=b&c=d", "example.com?a=b&c=d"),
            ("//example.com/?c=d&a=b", "example.com?a=b&c=d"),
            # query with quotes
            ("//example.com/?a=%22b%22", 'example.com?a="b"'),
            ("//example.com/?a=%27b%27", "example.com?a='b'"),
            # complext query with duplicates and sorting
            ("//example.com/?c=d&a=g&e=f&a=b", "example.com?a=b&a=g&c=d&e=f"),
            # duble slashes
            ("//example.com//path//to//resource", "example.com/path/to/resource"),
            # unicode normalization
            ("//example.com/%D0%BF%D1%80%D0%B8%D0%B2%D0%B5%D1%82", "example.com/привет"),
            # unicode complex symbols
            ("//example.com/%E2%98%83", "example.com/☃"),
            # unicode NFD to NFC
            ("//example.com/й", "example.com/й"),
            # remove schema
            ("http://example.com", "example.com"),
            ("https://example.com", "example.com"),
            ("ftp://example.com", "example.com"),
            ("//example.com", "example.com"),
            # remove ports
            ("//example.com:666", "example.com"),
            ("//example.com:666/path", "example.com/path"),
            ("//example.com:80", "example.com"),
            ("//example.com:443/path", "example.com/path"),
            # remove fragment
            ("//example.com#fragment", "example.com"),
            ("//example.com/path#fragment", "example.com/path"),
            ("//example.com/path#fragment?query", "example.com/path"),
            ("//example.com/path?query#", "example.com/path?query"),
        ],
    )
    def test(self, url: AbsoluteUrl, uid: UrlUid) -> None:
        assert urls.url_to_uid(url) == uid


class TestUrlToSourceUid:

    # This parameters are a bit excessive
    # Current implementation of url_to_uid expects already normalized AbsoluteUrl
    @pytest.mark.parametrize(
        "url, source_uid",
        [
            ("   //example.com   ", "example.com"),
            ("//example.com/path/a/b?x=y", "example.com"),
            ("//ExamPle.com", "example.com"),
            ("//www.example.com", "example.com"),
            ("//subdomain.example.com", "subdomain.example.com"),
            ("//old.reddit.com", "reddit.com"),
            ("//api.reddit.com", "reddit.com"),
            ("http://api.reddit.com", "reddit.com"),
            ("https://api.reddit.com", "reddit.com"),
            ("ftp://api.reddit.com", "reddit.com"),
            ("//programming.reddit.com", "reddit.com"),
            ("//anotherreddit.com", "anotherreddit.com"),
            ("//xxx.anotherreddit.com", "xxx.anotherreddit.com"),
            # unicode normalization
            ("//фвыа.com", "фвыа.com"),
        ],
    )
    def test(self, url: AbsoluteUrl, source_uid: SourceUid) -> None:
        assert urls.url_to_source_uid(url) == source_uid


class TestCheckFurlError:

    def test_invalid_port(self) -> None:
        with pytest.raises(errors.ExpectedFUrlError):
            with urls.check_furl_error():
                furl("https://example.com:666666")

    def test_invalid_host(self) -> None:
        with pytest.raises(errors.ExpectedFUrlError):
            with urls.check_furl_error():
                furl("https://visit%20borkensite.com%20for%20test%20or%not%21/")

    def test_another_error(self) -> None:
        with pytest.raises(ZeroDivisionError):
            with urls.check_furl_error():
                1 / 0


class TestSchemaSupported:

    @pytest.mark.parametrize(
        "schema, supported", [("https", True), ("http", True), (None, True), ("newsletter", False)]
    )
    def test(self, schema: str | None, supported: bool) -> None:
        assert urls.schema_supported(schema) == supported, f"Expected {supported} for schema {schema}"

    @pytest.mark.parametrize(
        "url, schema",
        [
            ("https://example.com", "https"),
            ("http://example.com", "http"),
            ("example.com", None),
            ("newsletter:666:noreply@example.com", "newsletter"),
        ],
    )
    def test_furl_detects_schema(self, url: UnknownUrl, schema: str | None) -> None:
        f_url = urls.construct_f_url(url)
        assert f_url is not None
        assert f_url.scheme == schema, f"Expected {schema}, got {f_url.scheme} for {url}"


class TestFixClassicUrlToAbsolute:

    def test_no_dot_in_domain(self) -> None:
        assert urls._fix_classic_url_to_absolute("localhost") is None
        assert urls._fix_classic_url_to_absolute("examplecom") is None

    def test_no_public_suffix(self) -> None:
        assert urls._fix_classic_url_to_absolute("example") is None
        assert urls._fix_classic_url_to_absolute("example.myprivatezone") is None

    def test_can_not_construct_furl(self) -> None:
        assert urls._fix_classic_url_to_absolute("example.com:666666") is None

    def test_ok(self) -> None:
        assert urls._fix_classic_url_to_absolute("example.com") == "//example.com"
        assert urls._fix_classic_url_to_absolute("example.com/path") == "//example.com/path"
        assert urls._fix_classic_url_to_absolute("example.com/path?x=y") == "//example.com/path?x=y"
        assert urls._fix_classic_url_to_absolute("example.com/path?x=y#fragment") == "//example.com/path?x=y#fragment"


class TestNormalizeClassicUnknownUrl:

    def test_general_sceme(self) -> None:
        assert urls.normalize_classic_unknown_url(UnknownUrl("//example.com?")) == "//example.com"

    def test_relative_same_level(self) -> None:
        assert urls.normalize_classic_unknown_url(UnknownUrl("./a/b?x=y")) is None

    def test_relative_upper_level(self) -> None:
        assert urls.normalize_classic_unknown_url(UnknownUrl("../a/b?x=y")) is None
        assert urls.normalize_classic_unknown_url(UnknownUrl("../../a/b?x=y")) is None

    def test_furl_error(self) -> None:
        assert urls.normalize_classic_unknown_url(UnknownUrl("//example.com:9999999?")) is None

    def test_has_scheme(self) -> None:
        assert urls.normalize_classic_unknown_url(UnknownUrl("http://example.com?")) == "http://example.com"
        assert urls.normalize_classic_unknown_url(UnknownUrl("https://example.com?")) == "https://example.com"

    def test_no_domain(self) -> None:
        assert urls.normalize_classic_unknown_url(UnknownUrl("a/b?x=y")) is None
        assert urls.normalize_classic_unknown_url(UnknownUrl("localhost/x/y?a=b")) is None

    def test_ok(self) -> None:
        assert urls.normalize_classic_unknown_url(UnknownUrl("example.com/a/b/c?x=y")) == "//example.com/a/b/c?x=y"

    def test_keepsegment(self) -> None:
        # short logic path
        assert (
            urls.normalize_classic_unknown_url(UnknownUrl("//example.com/a/b/c?x=y#z")) == "//example.com/a/b/c?x=y#z"
        )

        # longest logic path
        assert urls.normalize_classic_unknown_url(UnknownUrl("example.com/a/b/c?x=y#z")) == "//example.com/a/b/c?x=y#z"

    def remove_trailing_root_slash(self) -> None:
        assert urls.normalize_classic_unknown_url(UnknownUrl("example.com")) == "//example.com"
        assert urls.normalize_classic_unknown_url(UnknownUrl("example.com/")) == "//example.com"
        assert urls.normalize_classic_unknown_url(UnknownUrl("example.com//")) == "//example.com"
        assert urls.normalize_classic_unknown_url(UnknownUrl("example.com///")) == "//example.com"
        assert urls.normalize_classic_unknown_url(UnknownUrl("example.com////")) == "//example.com"

    def test_non_standard_schema(self) -> None:
        assert urls.normalize_classic_unknown_url(UnknownUrl("newsletter:666:noreply@example.com")) is None


class TestIsFullUrl:

    def test(self) -> None:
        assert urls.is_full_url(UnknownUrl("http://example.com"))
        assert not urls.is_full_url(UnknownUrl("abc"))


class TestStrToAbsoluteUrl:

    def test_ok(self) -> None:
        assert urls.str_to_absolute_url("http://example.com?") == "http://example.com"

    def test_not_ok(self) -> None:
        with pytest.raises(errors.UrlIsNotAbsolute):
            assert urls.str_to_absolute_url("example_com")


class TestStrToFeedUrl:

    def test_ok(self) -> None:
        assert urls.str_to_feed_url("http://example.com?") == "http://example.com"

    def test_not_ok(self) -> None:
        with pytest.raises(errors.UrlIsNotAbsolute):
            assert urls.str_to_feed_url("example_com")


class TestIsAbsoluteUrl:

    def test_ok(self) -> None:
        assert urls.is_absolute_url("http://example.com")

    def test_not_ok(self) -> None:
        assert not urls.is_absolute_url("http://example.com?")


class TestAdjustClassicFullUrl:

    def test_no_schema(self) -> None:
        assert (
            urls.adjust_classic_full_url(UnknownUrl("example.com"), urls.str_to_absolute_url("https://example.com"))
            == "https://example.com"
        )
        assert (
            urls.adjust_classic_full_url(UnknownUrl("example.com"), urls.str_to_absolute_url("http://example.com"))
            == "http://example.com"
        )

        assert (
            urls.adjust_classic_full_url(UnknownUrl("abc.com"), urls.str_to_absolute_url("http://example.com"))
            == "http://abc.com"
        )
        assert (
            urls.adjust_classic_full_url(UnknownUrl("abc.com"), urls.str_to_absolute_url("https://example.com"))
            == "https://abc.com"
        )

    def test_has_schema(self) -> None:
        assert (
            urls.adjust_classic_full_url(
                UnknownUrl("https://example.com"), urls.str_to_absolute_url("https://example.com")
            )
            == "https://example.com"
        )
        assert (
            urls.adjust_classic_full_url(
                UnknownUrl("https://example.com"), urls.str_to_absolute_url("http://example.com")
            )
            == "https://example.com"
        )

        assert (
            urls.adjust_classic_full_url(UnknownUrl("http://abc.com"), urls.str_to_absolute_url("http://example.com"))
            == "http://abc.com"
        )
        assert (
            urls.adjust_classic_full_url(UnknownUrl("http://abc.com"), urls.str_to_absolute_url("https://example.com"))
            == "http://abc.com"
        )


class TestAdjustClassicRelativeUrl:

    def test_cleanup_original(self) -> None:
        assert (
            urls.adjust_classic_relative_url(UnknownUrl("a/b"), urls.str_to_absolute_url("https://example.com?x=y#z"))
            == "https://example.com/a/b"
        )

    def test_same_level(self) -> None:
        assert (
            urls.adjust_classic_relative_url(
                UnknownUrl("./a/b"), urls.str_to_absolute_url("https://example.com/feed/part?x=y#z")
            )
            == "https://example.com/feed/a/b"
        )

    def test_upper_level(self) -> None:
        assert (
            urls.adjust_classic_relative_url(
                UnknownUrl("../a/b"), urls.str_to_absolute_url("https://example.com/feed/part?x=y#z")
            )
            == "https://example.com/a/b"
        )


class TestUrlHasExtension:

    def test_no_path(self) -> None:
        assert not urls.url_has_extension(urls.str_to_absolute_url("https://example.com/"), [".xml"])
        assert not urls.url_has_extension(urls.str_to_absolute_url("https://example.com"), [".xml"])
        assert urls.url_has_extension(urls.str_to_absolute_url("https://example.com/"), [""])
        assert urls.url_has_extension(urls.str_to_absolute_url("https://example.com"), [""])
        assert not urls.url_has_extension(urls.str_to_absolute_url("https://example.com/"), [])
        assert not urls.url_has_extension(urls.str_to_absolute_url("https://example.com"), [])

    def test_has(self) -> None:
        assert urls.url_has_extension(urls.str_to_absolute_url("https://example.com/feed.xml"), [".xml"])
        assert urls.url_has_extension(urls.str_to_absolute_url("https://example.com/feed.second..xml"), [".xml"])
        assert urls.url_has_extension(
            urls.str_to_absolute_url("https://example.com/feed.second.xml"), [".rss", ".xml", ".html"]
        )
        assert urls.url_has_extension(urls.str_to_absolute_url("https://example.com/feed/second"), [""])

    def test_has_no(self) -> None:
        assert not urls.url_has_extension(urls.str_to_absolute_url("https://example.com/feed.xml"), [])
        assert not urls.url_has_extension(urls.str_to_absolute_url("https://example.com/feed.second..xml"), [])
        assert not urls.url_has_extension(
            urls.str_to_absolute_url("https://example.com/feed.second.xml"), [".rss", ".html"]
        )
        assert not urls.url_has_extension(urls.str_to_absolute_url("https://example.com/feed/second"), [])


class TestFilterOutDuplicatedUrls:

    def test_no_urls(self) -> None:
        assert urls.filter_out_duplicated_urls([]) == []

    def test_no_duplicates(self) -> None:
        assert urls.filter_out_duplicated_urls(
            [
                urls.str_to_absolute_url("https://example.com/feed1"),
                urls.str_to_absolute_url("https://example.com/feed2"),
            ]
        ) == [
            "https://example.com/feed1",
            "https://example.com/feed2",
        ]

    def test_duplicates(self) -> None:
        assert urls.filter_out_duplicated_urls(
            [
                urls.str_to_absolute_url("https://example.com/feed1?a=b&c=d"),
                urls.str_to_absolute_url("https://example.com/feed2"),
                urls.str_to_absolute_url("https://example.com/feed1?c=d&a=b"),
            ]
        ) == ["https://example.com/feed1?a=b&c=d", "https://example.com/feed2"]


class TestGetParentUrl:

    def test_no_parent(self) -> None:
        assert urls.get_parent_url(urls.str_to_absolute_url("https://example.com")) is None
        assert urls.get_parent_url(urls.str_to_absolute_url("https://example.com/")) is None
        assert urls.get_parent_url(urls.str_to_absolute_url("https://subdomain.example.com")) is None

    def test_has_parent(self) -> None:
        assert urls.get_parent_url(urls.str_to_absolute_url("https://example.com/feed")) == "https://example.com"
        assert urls.get_parent_url(urls.str_to_absolute_url("https://example.com/feed/")) == "https://example.com/feed"
        assert (
            urls.get_parent_url(urls.str_to_absolute_url("https://example.com/feed/part"))
            == "https://example.com/feed/"
        )
        assert urls.get_parent_url(urls.str_to_absolute_url("https://example.com/feed/")) == "https://example.com/feed"
        assert (
            urls.get_parent_url(urls.str_to_absolute_url("https://subdomain.example.com/feed/part"))
            == "https://subdomain.example.com/feed/"
        )
        assert (
            urls.get_parent_url(urls.str_to_absolute_url("https://example.com/feed/part/second"))
            == "https://example.com/feed/part/"
        )


class TestToFeedUrl:

    @pytest.mark.parametrize(
        "url, feed_url",
        [
            ("https://example.com/feed", "https://example.com/feed"),
            ("https://example.com/feed/", "https://example.com/feed/"),
            ("https://example.com/feed#fragment", "https://example.com/feed"),
        ],
    )
    def test(self, url: AbsoluteUrl, feed_url: FeedUrl) -> None:
        assert urls.to_feed_url(url) == feed_url



"""
migrate_to_processors_queue
"""

import uuid
from typing import Any

from psycopg import Connection
from psycopg.rows import dict_row
from pypika import PostgreSQLQuery
from yoyo import step

__depends__ = {"20240313_01_Yv74T-processors-pointers", "20240315_01_tWftt-optimize-l-entries"}


def get_actual_entries(cursor: Any) -> set[uuid.UUID]:
    cursor.execute("SELECT id FROM l_entries")

    return {row["id"] for row in cursor.fetchall()}


def get_old_entries(cursor: Any, processor_id: int, state: int) -> list[uuid.UUID]:
    cursor.execute(
        "SELECT entry_id FROM l_entry_process_info WHERE processor_id = %(processor_id)s AND state = %(state)s",
        {"processor_id": processor_id, "state": state},
    )

    return [row["entry_id"] for row in cursor.fetchall()]


def add_entries_to_failed_storage(cursor: Any, processor_id: int, failed_entries: list[uuid.UUID]) -> None:
    chunk = 10000

    for i in range(0, len(failed_entries), chunk):
        query = PostgreSQLQuery.into("ln_failed_entries").columns("processor_id", "entry_id")

        for entry_id in failed_entries[i : i + chunk]:
            query = query.insert((processor_id, entry_id))

        cursor.execute(str(query))


def add_entries_to_queue(cursor: Any, processor_id: int, entries: list[uuid.UUID]) -> None:
    chunk = 10000

    for i in range(0, len(entries), chunk):
        query = PostgreSQLQuery.into("ln_processors_queue").columns("processor_id", "entry_id")

        for entry_id in entries[i : i + chunk]:
            query = query.insert((processor_id, entry_id))

        cursor.execute(str(query))


def create_processor_pointer(cursor: Any, processor_id: int) -> None:
    sql = """
    INSERT INTO ln_processor_pointers (processor_id, pointer_created_at, pointer_entry_id)
    VALUES (%(processor_id)s, NOW(), '00000000-0000-0000-0000-000000000000')
    """
    cursor.execute(sql, {"processor_id": processor_id})


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor(row_factory=dict_row)

    all_entries = get_actual_entries(cursor)

    for processor_id in (1, 2, 3, 4, 5):
        failed_entries = get_old_entries(cursor, processor_id, 2)
        add_entries_to_failed_storage(cursor, processor_id, failed_entries)

        successed_entries = get_old_entries(cursor, processor_id, 1)
        retried_entries = get_old_entries(cursor, processor_id, 3)

        entries_to_add = all_entries - set(successed_entries) - set(retried_entries) - set(failed_entries)
        add_entries_to_queue(cursor, processor_id, list(entries_to_add))

        create_processor_pointer(cursor, processor_id)


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    # the l_entry_process_info will be removed in https://github.com/Tiendil/feeds.fun/issues/177

    cursor = conn.cursor()

    cursor.execute("DELETE FROM ln_processors_queue")
    cursor.execute("DELETE FROM ln_failed_entries")
    cursor.execute("DELETE FROM ln_processor_pointers")


steps = [step(apply_step, rollback_step)]



import uuid
from decimal import Decimal

import pytest
import pytest_asyncio

from ffun.domain.datetime_intervals import month_interval_start
from ffun.domain.entities import UserId
from ffun.llms_framework.entities import LLMApiKey, LLMProvider, USDCost, UserKeyInfo
from ffun.llms_framework.keys_rotator import _cost_points, _get_user_key_infos
from ffun.llms_framework.keys_statuses import Statuses
from ffun.llms_framework.provider_interface import ProviderTest
from ffun.resources import domain as r_domain
from ffun.user_settings import domain as us_domain


# TODO: replace LLMProvider.openai with LLMProvider.test
@pytest_asyncio.fixture
async def five_user_key_infos(
    fake_llm_provider: ProviderTest, five_internal_user_ids: list[UserId]
) -> list[UserKeyInfo]:
    from ffun.application.resources import Resource as AppResource
    from ffun.application.user_settings import UserSetting

    max_tokens_cost_in_month = USDCost(Decimal(1000))
    used_cost = USDCost(Decimal(345))

    interval_started_at = month_interval_start()

    for user_id in five_internal_user_ids:
        await us_domain.save_setting(user_id=user_id, kind=UserSetting.test_api_key, value=uuid.uuid4().hex)

        await us_domain.save_setting(
            user_id=user_id, kind=UserSetting.max_tokens_cost_in_month, value=max_tokens_cost_in_month
        )

        await us_domain.save_setting(user_id=user_id, kind=UserSetting.process_entries_not_older_than, value=3)

        await r_domain.try_to_reserve(
            user_id=user_id,
            kind=AppResource.tokens_cost,
            interval_started_at=interval_started_at,
            amount=_cost_points.to_points(used_cost),
            limit=_cost_points.to_points(max_tokens_cost_in_month),
        )

        await r_domain.convert_reserved_to_used(
            user_id=user_id,
            kind=AppResource.tokens_cost,
            interval_started_at=interval_started_at,
            used=_cost_points.to_points(used_cost),
            reserved=_cost_points.to_points(used_cost),
        )

    return await _get_user_key_infos(LLMProvider.test, five_internal_user_ids, interval_started_at)


@pytest.fixture
def fake_llm_api_key() -> LLMApiKey:
    return LLMApiKey(uuid.uuid4().hex)


@pytest.fixture
def api_key_statuses() -> Statuses:
    return Statuses()


@pytest.fixture
def fake_llm_provider() -> ProviderTest:
    return ProviderTest()



import datetime
import enum

import pydantic

from ffun.core import utils
from ffun.core.entities import BaseEntity
from ffun.domain.entities import TagId


class TagPropertyType(int, enum.Enum):
    link = 2
    categories = 3


class TagCategory(str, enum.Enum):
    network_domain = "network-domain"
    feed_tag = "feed-tag"


class TagProperty(BaseEntity):
    tag_id: TagId
    type: TagPropertyType
    value: str
    processor_id: int
    created_at: datetime.datetime


class ProcessorTag(pydantic.BaseModel):
    raw_uid: str

    name: str | None = None
    link: str | None = None
    categories: set[TagCategory] = pydantic.Field(default_factory=set)

    def build_properties_for(self, tag_id: TagId, processor_id: int) -> list[TagProperty]:
        properties = []

        created_at = utils.now()

        if self.link:
            properties.append(
                TagProperty(
                    tag_id=tag_id,
                    type=TagPropertyType.link,
                    value=self.link,
                    processor_id=processor_id,
                    created_at=created_at,
                )
            )

        if self.categories:
            categories_dump = ",".join(sorted(c.value for c in self.categories))
            properties.append(
                TagProperty(
                    tag_id=tag_id,
                    type=TagPropertyType.categories,
                    value=categories_dump,
                    processor_id=processor_id,
                    created_at=created_at,
                )
            )

        return properties


class Tag(pydantic.BaseModel):
    id: int
    name: str | None = None
    link: str | None = None
    categories: set[TagCategory] = pydantic.Field(default_factory=set)



import enum

import pydantic
import pydantic_settings

from ffun.core.settings import BaseSettings


class AuthMode(str, enum.Enum):
    single_user = "single_user"
    supertokens = "supertokens"


class SingleUser(pydantic.BaseModel):
    external_id: str = "user-for-development"


class Supertokens(pydantic.BaseModel):
    connection_uri: str = "http://supertokens:3567"
    api_key: str = "nn4PGU5rJ3tEe9if4zEJ"  # this is a fake key for tests
    cookie_secure: bool = False

    api_base_path: str = "/supertokens"
    website_base_path: str = "/auth"


class Settings(BaseSettings):
    mode: AuthMode = AuthMode.single_user
    single_user: SingleUser = SingleUser()
    supertokens: Supertokens = Supertokens()

    model_config = pydantic_settings.SettingsConfigDict(env_prefix="FFUN_AUTH_")


settings = Settings()



import uuid
from unittest.mock import AsyncMock, MagicMock

import fastapi
import pytest
from httpx import AsyncClient
from pytest_mock import MockerFixture

from ffun.core import errors, json, logging
from ffun.core.middlewares import (
    _existed_route_urls,
    _handle_api_error,
    _handle_unexpected_error,
    _normalize_path,
    request_id_middleware,
)
from ffun.core.tests.helpers import capture_logs

logger = logging.get_module_logger()


class FakeError(errors.Error):
    pass


class TestHandleAPIError:

    @pytest.mark.asyncio
    async def test(self, mocker: MockerFixture) -> None:
        capture_exception = mocker.patch("sentry_sdk.capture_exception")

        error = errors.APIError(code="some.error.code", message="some message")

        request = MagicMock()

        with capture_logs() as logs:
            response = await _handle_api_error(request, error)

        assert request.state.api_error_code == "some.error.code"

        capture_exception.assert_not_called()

        assert logs == [
            {
                "event": "api_error",
                "log_level": "info",
                "module": "ffun.core.middlewares",
                "code": "some.error.code",
                "message": "some message",
            }
        ]

        assert response.status_code == 200

        body = response.body

        assert isinstance(body, bytes)

        assert json.parse(body.decode()) == {
            "status": "error",
            "code": "some.error.code",
            "message": "some message",
            "data": None,
        }


class TestHandleUnexpectedError:

    @pytest.mark.asyncio
    async def test(self, mocker: MockerFixture) -> None:
        capture_exception = mocker.patch("sentry_sdk.capture_exception")

        error = FakeError(message="some message")

        request = MagicMock()

        with capture_logs() as logs:
            response = await _handle_unexpected_error(request, error)

        assert request.state.internal_error_code == "FakeError"

        capture_exception.assert_called_once_with(error)

        assert logs == [
            {
                "event": "FakeError",
                "log_level": "error",
                "exc_info": True,
                "module": "ffun.core.middlewares",
                "sentry_skip": True,
            }
        ]

        assert response.status_code == 500

        body = response.body

        assert isinstance(body, bytes)

        assert json.parse(body.decode()) == {
            "status": "error",
            "code": "FakeError",
            "message": "An unexpected error appeared. We are working on fixing it.",
            "data": None,
        }


class TestRequestIdMiddleware:

    @pytest.mark.asyncio
    async def test(self, mocker: MockerFixture) -> None:
        bound_log_args = mocker.patch("ffun.core.logging.bound_log_args")

        await request_id_middleware(MagicMock(), AsyncMock())
        logger.info("request_id_test")

        assert bound_log_args.call_count == 1

        call = bound_log_args.call_args_list[0]

        assert uuid.UUID(call[1]["request_uid"])

    @pytest.mark.asyncio
    async def test_uniqueness(self, mocker: MockerFixture) -> None:
        bound_log_args = mocker.patch("ffun.core.logging.bound_log_args")

        await request_id_middleware(MagicMock(), AsyncMock())
        await request_id_middleware(MagicMock(), AsyncMock())

        logger.info("request_id_test")

        assert bound_log_args.call_count == 2

        call_1 = bound_log_args.call_args_list[0]
        call_2 = bound_log_args.call_args_list[1]

        assert call_1[1]["request_uid"] != call_2[1]["request_uid"]


class TestNormalizePath:

    @pytest.mark.parametrize(
        "url, expected",
        [
            ("/test", "/test"),
            ("/test/", "/test"),
            ("/test//", "/test"),
            ("/test/b", "/test/b"),
            ("/test/b/", "/test/b"),
            ("/test/b//", "/test/b"),
            ("/test//test", None),
            ("/test//test/", None),
        ],
    )
    def test(self, url: str, expected: str) -> None:
        assert _normalize_path(url) == expected


class TestExistedRouteUrls:

    @pytest.mark.asyncio
    async def test(self, app: fastapi.FastAPI) -> None:
        assert _existed_route_urls(app) == {
            "/api/get-rules",
            "/api/get-collections",
            "/api/add-opml",
            "/api/get-opml",
            "/api/delete-rule",
            "/api/add-feed",
            "/api/unsubscribe",
            "/api/docs",
            "/api/get-entries-by-ids",
            "/api/get-last-entries",
            "/api/get-last-collection-entries",
            "/api/subscribe-to-collections",
            "/api/test/ok",
            "/api/set-user-setting",
            "/api/test/internal-error",
            "/api/test/expected-error",
            "/api/get-collection-feeds",
            "/api/set-marker",
            "/api/get-score-details",
            "/api/discover-feeds",
            "/api/get-user-settings",
            "/api/get-resource-history",
            "/api/create-or-update-rule",
            "/api/get-feeds",
            "/api/get-info",
            "/api/get-tags-info",
            "/api/update-rule",
            "/api/openapi.json",
            "/api/remove-marker",
            "/api/track-event",
            "/api/remove-user",
        }


class TestRequestMeasureMiddleware:

    @pytest.mark.asyncio
    async def test(self, client: AsyncClient) -> None:
        with capture_logs() as logs:
            await client.post("/api/test/ok")

        logs = [record for record in logs if record["event"] == "request_time"]

        assert logs == [
            {
                "module": "ffun.core.middlewares",
                "m_kind": "measure",
                "m_value": logs[0]["m_value"],
                "event": "request_time",
                "request_uid": logs[0]["request_uid"],
                "m_labels": {"http_path": "/api/test/ok", "result": "success", "status_code": 200, "error_code": None},
                "log_level": "info",
            }
        ]

    @pytest.mark.asyncio
    async def test_other_status_code(self, client: AsyncClient) -> None:
        with capture_logs() as logs:
            await client.get("/api/test/ok")

        logs = [record for record in logs if record["event"] == "request_time"]

        assert logs == [
            {
                "module": "ffun.core.middlewares",
                "m_kind": "measure",
                "m_value": logs[0]["m_value"],
                "event": "request_time",
                "request_uid": logs[0]["request_uid"],
                "m_labels": {"http_path": "/api/test/ok", "result": "success", "status_code": 405, "error_code": None},
                "log_level": "info",
            }
        ]

    @pytest.mark.asyncio
    async def test_wrong_url(self, client: AsyncClient) -> None:
        with capture_logs() as logs:
            await client.post("/api/bla-bla-bla")

        logs = [record for record in logs if record["event"] == "request_time"]

        assert logs == [
            {
                "module": "ffun.core.middlewares",
                "m_kind": "measure",
                "m_value": logs[0]["m_value"],
                "event": "request_time",
                "request_uid": logs[0]["request_uid"],
                "m_labels": {"http_path": "wrong", "result": "success", "status_code": 404, "error_code": None},
                "log_level": "info",
            }
        ]

    @pytest.mark.asyncio
    async def test_internal_error(self, client: AsyncClient) -> None:
        with capture_logs() as logs:
            await client.post("/api/test/internal-error")

        logs = [record for record in logs if record["event"] == "request_time"]

        assert logs[0] == {
            "module": "ffun.core.middlewares",
            "m_kind": "measure",
            "m_value": logs[0]["m_value"],
            "event": "request_time",
            "request_uid": logs[0]["request_uid"],
            "m_labels": {
                "http_path": "/api/test/internal-error",
                "result": "internal_error",
                "status_code": 500,
                "error_code": "Exception",
            },
            "log_level": "info",
        }

    @pytest.mark.asyncio
    async def test_expected_error(self, client: AsyncClient) -> None:
        with capture_logs() as logs:
            await client.post("/api/test/expected-error")

        logs = [record for record in logs if record["event"] == "request_time"]

        assert logs[0] == {
            "module": "ffun.core.middlewares",
            "m_kind": "measure",
            "m_value": logs[0]["m_value"],
            "event": "request_time",
            "request_uid": logs[0]["request_uid"],
            "m_labels": {
                "http_path": "/api/test/expected-error",
                "result": "api_error",
                "status_code": 200,
                "error_code": "expected_test_error",
            },
            "log_level": "info",
        }



import pytest_asyncio

from ffun.feeds.domain import get_feed, mark_feed_as_loaded, save_feed
from ffun.feeds.entities import Feed, FeedId
from ffun.feeds.tests import make as f_make


@pytest_asyncio.fixture
async def new_feed() -> Feed:
    return await f_make.fake_feed()


@pytest_asyncio.fixture
async def another_new_feed() -> Feed:
    return await f_make.fake_feed()


@pytest_asyncio.fixture
async def saved_feed_id(new_feed: Feed) -> FeedId:
    return await save_feed(new_feed)


@pytest_asyncio.fixture
async def another_saved_feed_id(another_new_feed: Feed) -> FeedId:
    return await save_feed(another_new_feed)


@pytest_asyncio.fixture
async def saved_feed(saved_feed_id: FeedId) -> Feed:
    return await get_feed(saved_feed_id)


@pytest_asyncio.fixture
async def another_saved_feed(another_saved_feed_id: FeedId) -> Feed:
    return await get_feed(another_saved_feed_id)


@pytest_asyncio.fixture
async def loaded_feed_id(saved_feed_id: FeedId) -> FeedId:
    await mark_feed_as_loaded(saved_feed_id)
    return saved_feed_id


@pytest_asyncio.fixture
async def another_loaded_feed_id(another_saved_feed_id: FeedId) -> FeedId:
    await mark_feed_as_loaded(another_saved_feed_id)
    return another_saved_feed_id


@pytest_asyncio.fixture
async def loaded_feed(loaded_feed_id: FeedId) -> Feed:
    return await get_feed(loaded_feed_id)


@pytest_asyncio.fixture
async def another_loaded_feed(another_loaded_feed_id: FeedId) -> Feed:
    return await get_feed(another_loaded_feed_id)


@pytest_asyncio.fixture
async def five_saved_feed_ids() -> list[FeedId]:
    return [feed.id for feed in await f_make.n_feeds(5)]



from typing import Any, Iterable

import psycopg
from pypika import PostgreSQLQuery

from ffun.core import logging
from ffun.core.postgresql import ExecuteType, execute
from ffun.domain.entities import EntryId
from ffun.librarian import errors
from ffun.librarian.entities import ProcessorPointer

logger = logging.get_module_logger()


def row_to_processor_pointer(row: dict[str, Any]) -> ProcessorPointer:
    return ProcessorPointer(
        processor_id=row["processor_id"],
        pointer_created_at=row["pointer_created_at"],
        pointer_entry_id=row["pointer_entry_id"],
    )


async def get_or_create_pointer(processor_id: int) -> ProcessorPointer:
    sql = """
    SELECT * FROM ln_processor_pointers
    WHERE processor_id = %(processor_id)s
    """

    row = await execute(sql, {"processor_id": processor_id})

    if row:
        return row_to_processor_pointer(row[0])

    created_pointer = await create_pointer(processor_id)

    if created_pointer is None:
        return await get_or_create_pointer(processor_id)

    return created_pointer


async def create_pointer(processor_id: int) -> ProcessorPointer | None:
    sql = """
    INSERT INTO ln_processor_pointers (processor_id)
    VALUES (%(processor_id)s)
    RETURNING *
    """

    try:
        row = await execute(sql, {"processor_id": processor_id})
    except psycopg.errors.UniqueViolation:
        return None

    return row_to_processor_pointer(row[0])


async def delete_pointer(processor_id: int) -> None:
    sql = """
    DELETE FROM ln_processor_pointers
    WHERE processor_id = %(processor_id)s
    """

    await execute(sql, {"processor_id": processor_id})


async def save_pointer(execute: ExecuteType, pointer: ProcessorPointer) -> None:
    sql = """
    UPDATE ln_processor_pointers
    SET pointer_created_at = %(pointer_created_at)s,
        pointer_entry_id = %(pointer_entry_id)s,
        updated_at = NOW()
    WHERE processor_id = %(processor_id)s
    RETURNING *
    """

    row = await execute(
        sql,
        {
            "processor_id": pointer.processor_id,
            "pointer_created_at": pointer.pointer_created_at,
            "pointer_entry_id": pointer.pointer_entry_id,
        },
    )

    if not row:
        raise errors.CanNotSaveUnexistingPointer()


async def get_all_pointers() -> list[ProcessorPointer]:
    sql = """
        SELECT processor_id, pointer_created_at, pointer_entry_id
        FROM ln_processor_pointers
    """

    rows = await execute(sql)

    return [row_to_processor_pointer(row) for row in rows]


async def push_entries_to_processor_queue(
    execute: ExecuteType, processor_id: int, entry_ids: Iterable[EntryId]
) -> None:
    query = PostgreSQLQuery.into("ln_processors_queue").columns("processor_id", "entry_id")

    for entry_id in entry_ids:
        query = query.insert(processor_id, entry_id)

    await execute(str(query))


async def get_entries_to_process(processor_id: int, limit: int) -> list[EntryId]:
    sql = """
    SELECT entry_id FROM ln_processors_queue
    WHERE processor_id = %(processor_id)s
    ORDER BY created_at ASC
    LIMIT %(limit)s
    """

    rows = await execute(sql, {"processor_id": processor_id, "limit": limit})

    return [row["entry_id"] for row in rows]


async def count_entries_in_processor_queue(processor_id: int) -> int:
    sql = """
    SELECT COUNT(*) FROM ln_processors_queue
    WHERE processor_id = %(processor_id)s
    """

    rows = await execute(sql, {"processor_id": processor_id})

    return rows[0]["count"]  # type: ignore


async def remove_entries_from_processor_queue(
    execute: ExecuteType, processor_id: int, entry_ids: Iterable[EntryId]
) -> None:
    sql = """
    DELETE FROM ln_processors_queue
    WHERE processor_id = %(processor_id)s
    AND entry_id = ANY(%(entry_ids)s)
    """

    await execute(sql, {"processor_id": processor_id, "entry_ids": list(entry_ids)})


async def clear_processor_queue(processor_id: int) -> None:
    sql = """
    DELETE FROM ln_processors_queue
    WHERE processor_id = %(processor_id)s
    """

    await execute(sql, {"processor_id": processor_id})


async def add_entries_to_failed_storage(processor_id: int, entry_ids: Iterable[EntryId]) -> None:
    query = PostgreSQLQuery.into("ln_failed_entries").columns("processor_id", "entry_id")

    for entry_id in entry_ids:
        query = query.insert(processor_id, entry_id)

    await execute(str(query))


async def get_failed_entries(execute: ExecuteType, processor_id: int, limit: int) -> list[EntryId]:
    sql = """
    SELECT entry_id FROM ln_failed_entries
    WHERE processor_id = %(processor_id)s
    ORDER BY created_at ASC
    LIMIT %(limit)s
    """

    rows = await execute(sql, {"processor_id": processor_id, "limit": limit})

    return [row["entry_id"] for row in rows]


async def remove_failed_entries(execute: ExecuteType, processor_id: int, entry_ids: Iterable[EntryId]) -> None:
    sql = """
    DELETE FROM ln_failed_entries
    WHERE processor_id = %(processor_id)s
    AND entry_id = ANY(%(entry_ids)s)
    """

    await execute(sql, {"processor_id": processor_id, "entry_ids": list(entry_ids)})


async def count_failed_entries() -> dict[int, int]:
    sql = """
    SELECT processor_id, COUNT(*) FROM ln_failed_entries GROUP BY processor_id
    """

    rows = await execute(sql)

    return {row["processor_id"]: row["count"] for row in rows}



import random
import uuid

import pytest

from ffun.core.tests.helpers import assert_logs_has_business_event, capture_logs
from ffun.domain.entities import UserId
from ffun.user_settings import operations
from ffun.user_settings.tests import asserts

_kind_1 = 1234
_kind_2 = 5678


class TestSave:
    @pytest.mark.asyncio
    async def test_save_load(self, internal_user_id: UserId) -> None:
        for value in ("abc", "def"):
            await operations.save_setting(internal_user_id, _kind_1, value)
            await asserts.has_settings(internal_user_id, {_kind_1: value})

    @pytest.mark.asyncio
    async def test_save_business_event(self, internal_user_id: UserId) -> None:
        with capture_logs() as logs:
            await operations.save_setting(internal_user_id, _kind_1, "xxx")

        assert_logs_has_business_event(logs, "setting_updated", user_id=internal_user_id, kind=_kind_1, first_set=True)

    @pytest.mark.asyncio
    async def test_update_on_existed_setting(self, internal_user_id: UserId) -> None:
        await operations.save_setting(internal_user_id, _kind_1, "xxx")

        with capture_logs() as logs:
            await operations.save_setting(internal_user_id, _kind_1, "yyy")

        assert_logs_has_business_event(
            logs, "setting_updated", user_id=internal_user_id, kind=_kind_1, first_set=False
        )

    @pytest.mark.asyncio
    async def test_save_load_multiple_kinds(self, internal_user_id: UserId) -> None:
        await operations.save_setting(internal_user_id, _kind_1, "abc")
        await operations.save_setting(internal_user_id, _kind_2, "def")

        await asserts.has_settings(internal_user_id, {_kind_1: "abc", _kind_2: "def"})

    @pytest.mark.asyncio
    async def test_save_load_multiple_users(self, internal_user_id: UserId, another_internal_user_id: UserId) -> None:
        await operations.save_setting(internal_user_id, _kind_1, "abc")
        await operations.save_setting(another_internal_user_id, _kind_1, "def")

        await asserts.has_settings(internal_user_id, {_kind_1: "abc"})
        await asserts.has_settings(another_internal_user_id, {_kind_1: "def"})


class TestLoadSettingsForUsers:
    @pytest.mark.asyncio
    async def test_no_users(self) -> None:
        assert await operations.load_settings_for_users([], [_kind_1]) == {}

    @pytest.mark.asyncio
    async def test_no_kinds(self, internal_user_id: UserId) -> None:
        assert await operations.load_settings_for_users([internal_user_id], []) == {internal_user_id: {}}

    @pytest.mark.asyncio
    async def test_no_settings_stored(self, internal_user_id: UserId, another_internal_user_id: UserId) -> None:
        settings = await operations.load_settings_for_users(
            [internal_user_id, another_internal_user_id], [_kind_1, _kind_2]
        )

        assert settings == {internal_user_id: {}, another_internal_user_id: {}}

    @pytest.mark.asyncio
    async def test_some_settings_stored(self, internal_user_id: UserId, another_internal_user_id: UserId) -> None:
        await operations.save_setting(internal_user_id, _kind_1, "abc")
        await operations.save_setting(internal_user_id, _kind_2, "def")
        await operations.save_setting(another_internal_user_id, _kind_1, "ghi")

        settings = await operations.load_settings_for_users(
            [internal_user_id, another_internal_user_id], [_kind_1, _kind_2]
        )

        assert settings == {
            internal_user_id: {_kind_1: "abc", _kind_2: "def"},
            another_internal_user_id: {_kind_1: "ghi"},
        }


class TestGetUsersWithSetting:
    @pytest.mark.asyncio
    async def test_no_users(self) -> None:
        assert await operations.get_users_with_setting(_kind_1, uuid.uuid4().hex) == set()

    @pytest.mark.asyncio
    async def test_ignore_value_with_wrong_kind(
        self, internal_user_id: UserId, another_internal_user_id: UserId
    ) -> None:
        value_1 = uuid.uuid4().hex

        await operations.save_setting(internal_user_id, _kind_1, value_1)
        await operations.save_setting(another_internal_user_id, _kind_2, value_1)

        assert await operations.get_users_with_setting(_kind_1, value_1) == {internal_user_id}
        assert await operations.get_users_with_setting(_kind_2, value_1) == {another_internal_user_id}

    @pytest.mark.asyncio
    async def test_has_users(self, internal_user_id: UserId, another_internal_user_id: UserId) -> None:
        value_1 = uuid.uuid4().hex
        value_2 = uuid.uuid4().hex
        value_3 = uuid.uuid4().hex

        await operations.save_setting(internal_user_id, _kind_1, value_1)
        await operations.save_setting(another_internal_user_id, _kind_1, value_2)

        await operations.save_setting(internal_user_id, _kind_2, value_2)
        await operations.save_setting(another_internal_user_id, _kind_2, value_2)

        assert await operations.get_users_with_setting(_kind_1, value_1) == {internal_user_id}
        assert await operations.get_users_with_setting(_kind_1, value_2) == {another_internal_user_id}
        assert await operations.get_users_with_setting(_kind_1, value_3) == set()

        assert await operations.get_users_with_setting(_kind_2, value_1) == set()
        assert await operations.get_users_with_setting(_kind_2, value_2) == {
            internal_user_id,
            another_internal_user_id,
        }
        assert await operations.get_users_with_setting(_kind_2, value_3) == set()


class TestRemoveSettingForAllUsers:
    @pytest.mark.asyncio
    async def test(self, five_internal_user_ids: list[UserId]) -> None:
        value = uuid.uuid4().hex

        await operations.save_setting(five_internal_user_ids[0], _kind_1, value)

        await operations.save_setting(five_internal_user_ids[1], _kind_1, value)
        await operations.save_setting(five_internal_user_ids[1], _kind_2, value)

        await operations.save_setting(five_internal_user_ids[2], _kind_2, value)

        await operations.remove_setting_for_all_users(_kind_1)

        settings = await operations.load_settings_for_users(five_internal_user_ids[:3], [_kind_1, _kind_2])

        assert settings == {
            five_internal_user_ids[0]: {},
            five_internal_user_ids[1]: {_kind_2: value},
            five_internal_user_ids[2]: {_kind_2: value},
        }


class TestFindAllKinds:

    @pytest.mark.asyncio
    async def test_test(self, internal_user_id: UserId) -> None:
        kinds_before = await operations.find_all_kinds()

        while new_kind := random.randint(1, 1000000):
            if new_kind not in kinds_before:
                break

        await operations.save_setting(internal_user_id, new_kind, "abc")

        kinds_after = await operations.find_all_kinds()

        assert kinds_after - kinds_before == {new_kind}



from typing import Sequence

from ffun.llms_framework import errors
from ffun.llms_framework.entities import ChatRequest, ChatResponse, KeyStatus, LLMConfiguration, LLMProvider, ModelInfo
from ffun.llms_framework.keys_statuses import Statuses
from ffun.llms_framework.settings import settings


class ProviderInterface:

    provider: LLMProvider = NotImplemented

    def __init__(self) -> None:
        self.api_keys_statuses = Statuses()

    def get_model(self, config: LLMConfiguration) -> ModelInfo:
        for m in settings.models:
            if m.provider != self.provider:
                continue

            if m.name == config.model:
                return m

        raise errors.ModelDoesNotFound(model=config.model)

    def prepare_requests(self, config: LLMConfiguration, text: str) -> Sequence[ChatRequest]:
        raise NotImplementedError("Must be implemented in a subclass")

    def estimate_tokens(self, config: LLMConfiguration, text: str) -> int:
        raise NotImplementedError("Must be implemented in a subclass")

    async def chat_request(self, config: LLMConfiguration, api_key: str, request: ChatRequest) -> ChatResponse:
        raise NotImplementedError("Must be implemented in a subclass")

    async def check_api_key(self, config: LLMConfiguration, api_key: str) -> KeyStatus:
        raise NotImplementedError("Must be implemented in a subclass")


class ChatRequestTest(ChatRequest):
    text: str


class ChatResponseTest(ChatResponse):
    content: str

    def response_content(self) -> str:
        return self.content

    def input_tokens(self) -> int:
        return len(self.content)

    # TODO: make it differ from input_tokens
    def output_tokens(self) -> int:
        return len(self.content)


class ProviderTest(ProviderInterface):
    provider = LLMProvider.test

    def prepare_requests(self, config: LLMConfiguration, text: str) -> Sequence[ChatRequestTest]:
        return [ChatRequestTest(text=text)]

    def estimate_tokens(self, config: LLMConfiguration, text: str) -> int:
        """Count each character as a token to simplify tests"""
        return len(text)

    async def chat_request(  # type: ignore
        self, config: LLMConfiguration, api_key: str, request: ChatRequestTest
    ) -> ChatResponseTest:
        if "raise TemporaryError" in request.text:
            raise errors.TemporaryError(message="error raise by a test request")

        return ChatResponseTest(content=request.text)

    async def check_api_key(self, config: LLMConfiguration, api_key: str) -> KeyStatus:
        return KeyStatus.works


provider_test = ProviderTest()



import pathlib
from typing import Any

import pydantic
import pydantic_settings

from ffun.core.settings import BaseSettings

_root = pathlib.Path(__file__).parent


class Settings(BaseSettings):
    collection_configs: pathlib.Path | None = None

    model_config = pydantic_settings.SettingsConfigDict(env_prefix="FFUN_FEEDS_COLLECTIONS_")

    @pydantic.field_validator("collection_configs", mode="before")
    @classmethod
    def detect_disabled_collection_configs(cls, v: Any) -> Any:
        if v == "":
            return None

        return v


settings = Settings()



"""
column-for-feed-unique-id
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__ = {"20230426_01_IiF5m-add-title-and-description"}


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()

    sql = "ALTER TABLE f_feeds ADD COLUMN uid TEXT DEFAULT NULL"

    cursor.execute(sql)


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()

    sql = "ALTER TABLE f_feeds DROP COLUMN uid"

    cursor.execute(sql)


steps = [step(apply_step, rollback_step)]



"""
tags-relationship-processor-tracking
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__ = {"20230617_01_XDdNG-rename-tags-names-to-uid"}

sql_create_relatons_processors_table = """
CREATE TABLE o_relations_processors (
    relation_id BIGINT REFERENCES o_relations(id),
    processor_id BIGINT NOT NULL,
    created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW(),
    PRIMARY KEY (relation_id, processor_id)
)
"""


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute(sql_create_relatons_processors_table)


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("DROP TABLE o_relations_processors")


steps = [step(apply_step, rollback_step)]



"""
source-for-entries
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__ = {"20240506_02_zQdSl-fix-unique-index-on-l-entries", "20240928_01_OEKOr-source-table"}


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()

    cursor.execute("ALTER TABLE l_entries ADD COLUMN source_id UUID NULL DEFAULT NULL")

    cursor.execute(
        "UPDATE l_entries SET source_id = f_feeds.source_id FROM f_feeds WHERE l_entries.feed_id = f_feeds.id"
    )

    cursor.execute("ALTER TABLE l_entries ALTER COLUMN source_id SET NOT NULL")


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()

    cursor.execute("ALTER TABLE l_entries DROP COLUMN source_id")


steps = [step(apply_step, rollback_step)]



from typing import Any


class Error(Exception):
    def __init__(self, **kwargs: Any) -> None:
        for key, value in kwargs.items():
            setattr(self, key, value)

        # TODO: send to Sentry
        if "fingerprint" not in kwargs:
            self.fingerprint = None

    def __str__(self) -> str:
        attributes = ", ".join(f"{key}={value}" for key, value in self.__dict__.items())

        return f"{self.__class__.__name__}: {attributes}"


class CoreError(Error):
    pass


class APIError(CoreError):
    def __init__(self, code: str, message: str) -> None:
        super().__init__(code=code, message=message)


class EntityAlreadyRegistered(CoreError):
    pass


class ReservedLogArguments(CoreError):
    pass


class DuplicatedMeasureLabels(CoreError):
    pass


class DuplicatedLogArguments(CoreError):
    pass



import asyncio
import uuid

import typer

from ffun.application.application import with_app
from ffun.core import logging
from ffun.domain.entities import CollectionId, FeedUrl, UnknownUrl
from ffun.domain.urls import normalize_classic_unknown_url, to_feed_url
from ffun.feeds_collections.collections import collections
from ffun.loader.domain import extract_feed_info

logger = logging.get_module_logger()

cli_app = typer.Typer()


async def _estimate_entries_per_day(feed_url: FeedUrl) -> float | None:
    logger.info("start_feed_extraction", feed_url=feed_url)

    feed_info = await extract_feed_info(feed_id=None, feed_url=feed_url)

    if feed_info is None:
        logger.info("feed_extraction_failed", feed_url=feed_url)
        return None

    if not feed_info.entries:
        logger.info("feed_has_no_entries", feed_url=feed_url)
        return None

    if len(feed_info.entries) == 1:
        logger.info("feed_has_only_one_entry", feed_url=feed_url)
        return None

    min_published_at = min(entry.published_at for entry in feed_info.entries)
    max_published_at = max(entry.published_at for entry in feed_info.entries)

    entries_per_day = len(feed_info.entries) / (max_published_at - min_published_at).total_seconds() * 86400

    logger.info("estimated_entries_for_feed_in_day", feed_url=feed_url, entries_per_day=entries_per_day)

    return entries_per_day


async def run_entries_per_day_for_feed(raw_feed_url: str) -> None:
    async with with_app():
        feed_url = normalize_classic_unknown_url(UnknownUrl(raw_feed_url))

        assert feed_url is not None

        entries_per_day = await _estimate_entries_per_day(to_feed_url(feed_url))

        logger.info("estimated_entries_for_feed_in_day", feed_url=feed_url, entries_per_day=entries_per_day)


async def run_entries_per_day_for_collection(collection_id: uuid.UUID) -> None:
    async with with_app():
        logger.info("start_collection_extraction", collection_id=collection_id)

        collection = collections.collection(CollectionId(collection_id))

        urls = {feed_info.url for feed_info in collection.feeds}

        total_entries_per_day = 0.0

        for url in urls:
            entries_per_day = await _estimate_entries_per_day(url)

            if entries_per_day is not None:
                total_entries_per_day += entries_per_day

        logger.info(
            "estimated_entries_for_feed_in_day", collection_id=collection_id, entries_per_day=total_entries_per_day
        )


@cli_app.command()
def entries_per_day_for_feed(feed_url: str) -> None:
    asyncio.run(run_entries_per_day_for_feed(feed_url))


@cli_app.command()
def entries_per_day_for_collection(collection_id: uuid.UUID) -> None:
    asyncio.run(run_entries_per_day_for_collection(collection_id))



import datetime
import enum
from typing import TYPE_CHECKING, Annotated, Any, Literal, Protocol

import pydantic
from pydantic_core import PydanticCustomError

from ffun.core import logging
from ffun.core.entities import BaseEntity
from ffun.domain.entities import EntryId
from ffun.llms_framework.entities import LLMCollectionApiKey, LLMConfiguration, LLMGeneralApiKey, LLMProvider

logger = logging.get_module_logger()


class TextCleaner(Protocol):
    def __call__(self, text: str) -> str:
        pass


class TagsExtractor(Protocol):
    def __call__(self, text: str) -> set[str]:
        pass


class ProcessorType(enum.StrEnum):
    fake = "fake"
    domain = "domain"
    native_tags = "native_tags"
    upper_case_title = "upper_case_title"
    llm_general = "llm_general"


class ProcessorPointer(BaseEntity):
    processor_id: int
    pointer_created_at: datetime.datetime
    pointer_entry_id: EntryId


class BaseProcessor(BaseEntity):
    id: int
    enabled: bool
    workers: int
    name: str
    type: ProcessorType
    allowed_for_collections: bool
    allowed_for_users: bool


class DomainProcessor(BaseProcessor):
    type: Literal[ProcessorType.domain] = ProcessorType.domain


class NativeTagsProcessor(BaseProcessor):
    type: Literal[ProcessorType.native_tags] = ProcessorType.native_tags


class UpperCaseTitleProcessor(BaseProcessor):
    type: Literal[ProcessorType.upper_case_title] = ProcessorType.upper_case_title


_general_key_warning = "I understand that setting this key may be DANGEROUS and COSTLY."


if TYPE_CHECKING:
    # TODO: fix after Pydantic learn how to process such parametrization at runtime
    PydanticTextCleaner = pydantic.ImportString[TextCleaner]
    PydanticTagsExtractor = pydantic.ImportString[TagsExtractor]
else:
    PydanticTextCleaner = pydantic.ImportString
    PydanticTagsExtractor = pydantic.ImportString


class LLMGeneralProcessor(BaseProcessor):

    # arbitrary_types_allowed is required for parametrization of pydantic.ImportString
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    type: Literal[ProcessorType.llm_general] = ProcessorType.llm_general

    # TODO: validate that template will render correctly
    entry_template: str

    text_cleaner: PydanticTextCleaner
    tags_extractor: PydanticTagsExtractor

    llm_provider: LLMProvider

    llm_config: LLMConfiguration

    collections_api_key: LLMCollectionApiKey | None = None
    general_api_key: LLMGeneralApiKey | None = None

    general_api_key_warning: str | None = None

    @pydantic.model_validator(mode="before")
    @classmethod
    def collections_api_key_none(cls, data: Any) -> Any:
        """Translate empty string to None because TOML does not support None value"""
        if data.get("collections_api_key") == "":
            data["collections_api_key"] = None

        return data

    @pydantic.model_validator(mode="after")
    def require_collections_or_general_key_if_collections_are_allowed(self) -> "LLMGeneralProcessor":
        if self.allowed_for_collections and self.collections_api_key is None and self.general_api_key is None:
            raise PydanticCustomError(
                "collections_or_general_key_required_for_collections",
                "Collections API key or General API key must be set "
                "if you want to use this processor to process feeds from collections. "
                "Feeds Fun must guarantee that users will not pay for processing feeds from collections.",
            )

        return self

    @pydantic.model_validator(mode="before")
    @classmethod
    def general_api_key_none(cls, data: Any) -> Any:
        """Translate empty string to None because TOML does not support None value"""
        if data.get("general_api_key") == "":
            data["general_api_key"] = None

        return data

    @pydantic.model_validator(mode="after")
    def general_api_key_warning_check(self) -> "LLMGeneralProcessor":
        if self.general_api_key is not None and self.general_api_key_warning != _general_key_warning:
            raise PydanticCustomError(
                "you_must_confirm_general_api_key_usage",
                "You must confirm that you understand the risks of using the General API key. "
                f"Set `general_api_key_warning` to the value '{_general_key_warning}'.",
            )

        return self


TagProcessor = Annotated[
    DomainProcessor | NativeTagsProcessor | UpperCaseTitleProcessor | LLMGeneralProcessor,
    pydantic.Field(discriminator="type"),
]


class ProcessorsConfig(pydantic.BaseModel):
    tag_processors: tuple[TagProcessor, ...]



import asyncio
import re

from bs4 import BeautifulSoup

from ffun.core import logging
from ffun.domain.entities import AbsoluteUrl, UnknownUrl
from ffun.domain.urls import (
    adjust_classic_url,
    construct_f_url,
    filter_out_duplicated_urls,
    get_parent_url,
    normalize_classic_unknown_url,
    to_feed_url,
    url_has_extension,
)
from ffun.feeds_discoverer.entities import Context, Discoverer, Result, Status
from ffun.loader import domain as lo_domain
from ffun.loader import errors as lo_errors
from ffun.parsers import entities as p_entities

logger = logging.get_module_logger()


async def _discover_adjust_url(context: Context) -> tuple[Context, Result | None]:

    logger.info("discovering_adjusting_url", raw_url=context.raw_url)

    absolute_url = normalize_classic_unknown_url(context.raw_url)

    logger.info("discovering_absolute_url", raw_url=context.raw_url, adjusted_url=absolute_url)

    if absolute_url is None:
        logger.info("discovering_incorrect_url")
        return context, Result(feeds=[], status=Status.incorrect_url)

    url = to_feed_url(absolute_url)

    logger.info("discovering_normalized_url", raw_url=context.raw_url, adjusted_url=url)

    return context.replace(url=url), None


async def _discover_load_url(context: Context) -> tuple[Context, Result | None]:
    assert context.url is not None

    logger.info("discovering_loading_content", url=context.url)

    try:
        response = await lo_domain.load_content_with_proxies(context.url)
        content = await lo_domain.decode_content(response)
    except lo_errors.LoadError:
        logger.info("can_not_access_content")
        return context, Result(feeds=[], status=Status.cannot_access_url)
    except Exception:
        logger.exception("unexpected_error_while_parsing_feed")
        return context, Result(feeds=[], status=Status.cannot_access_url)

    logger.info("discovering_content_loaded", url=context.url, content_size=len(content))

    return context.replace(content=content), None


async def _discover_extract_feed_info(context: Context) -> tuple[Context, Result | None]:
    assert context.url is not None
    assert context.content is not None

    logger.info("discovering_extracting_feed_info", url=context.url)

    try:
        feed_info = await lo_domain.parse_content(context.content, original_url=context.url)
    except lo_errors.LoadError:
        logger.info("feed_not_found_at_start_url")
        return context, None
    except Exception:
        logger.exception("unexpected_error_while_parsing_feed")
        return context, None

    logger.info("discovering_feed_info_extracted", feed_info=feed_info)

    return context, Result(feeds=[feed_info], status=Status.feeds_found)


async def _discover_create_soup(context: Context) -> tuple[Context, Result | None]:
    assert context.content is not None

    logger.info("discovering_creating_soup")

    try:
        soup = BeautifulSoup(context.content, "html.parser")
    except Exception:
        logger.exception("unexpected_error_while_parsing_html")
        return context, Result(feeds=[], status=Status.not_html)

    logger.info("discovering_soup_created")

    return context.replace(soup=soup), None


async def _discover_extract_feeds_from_links(context: Context) -> tuple[Context, Result | None]:
    assert context.url is not None
    assert context.soup is not None

    logger.info("discovering_extracting_feeds_from_links", url=context.url)

    links_to_check = set()

    for link in context.soup("link"):
        if link.has_attr("href"):
            if link.has_attr("rel") and any(
                rel in link["rel"] for rel in ["author", "help", "icon", "license", "pingback", "search", "stylesheet"]
            ):
                continue
            links_to_check.add(adjust_classic_url(link["href"], context.url))

    logger.info("discovering_links_extracted", links_to_check=links_to_check)

    return context.replace(candidate_urls=context.candidate_urls | links_to_check), None


async def _discover_extract_feeds_from_anchors(context: Context) -> tuple[Context, Result | None]:
    assert context.url is not None

    logger.info("discovering_extracting_feeds_from_anchors", url=context.url)

    allowed_extensions = [".xml", ".rss", ".atom", ".rdf", ".feed", ".php", ".asp", ".aspx", ".json", ".cgi", ""]

    assert context.soup is not None

    anchors_to_check = set()

    for anchor in list(context.soup("a")):
        if anchor.has_attr("href"):
            url = anchor["href"]

            if not url_has_extension(url, allowed_extensions):
                continue

            anchors_to_check.add(adjust_classic_url(url, context.url))

    logger.info("discovering_anchors_extracted", anchors_to_check=anchors_to_check)

    return context.replace(candidate_urls=context.candidate_urls | anchors_to_check), None


async def _discover_check_parent_urls(context: Context) -> tuple[Context, Result | None]:
    assert context.url is not None

    logger.info("discovering_checking_parent_urls", url=context.url)

    parent_url = get_parent_url(context.url)

    if parent_url is None:
        logger.info("discovering_no_parent")
        return context, None

    # always check parents on depth of 1 (a.k.a., we expect html with links to feeds)
    result = await discover(url=parent_url, depth=1, discoverers=context.discoverers)

    if result.feeds:
        logger.info("discovering_parent_extracted", parent_url=parent_url)
        return context, result

    logger.info("discovering_no_parent_extracted")
    return context, None


async def _discover_check_candidate_links(context: Context) -> tuple[Context, Result | None]:  # noqa: CCR001

    logger.info("discovering_checking_links", candidate_links=context.candidate_urls)

    filtered_links = filter_out_duplicated_urls(context.candidate_urls)

    tasks = []

    for link in filtered_links:
        tasks.append(discover(url=link, depth=context.depth - 1, discoverers=context.discoverers))

    results: list[Result] = await asyncio.gather(*tasks)

    feeds: list[p_entities.FeedInfo] = []

    for result in results:
        feeds.extend(result.feeds)

    feeds.sort(key=lambda feed_info: feed_info.title)

    if not feeds:
        logger.info("discovering_checking_links_no_feeds_found")
        return context.replace(candidate_urls=set()), None

    logger.info("discovering_checking_links_feeds_found", result_links=[feed_info.url for feed_info in feeds])

    return context.replace(candidate_urls=set()), Result(feeds=feeds, status=Status.feeds_found)


async def _discover_stop_recursion(context: Context) -> tuple[Context, Result | None]:

    logger.info("discovering_check_recursion", depth=context.depth)

    if context.depth == 0:
        logger.info("discovering_recursion_stopped")
        return context, Result(feeds=[], status=Status.no_feeds_found)

    logger.info("discovering_recursion_continued")

    return context, None


_RE_REDDIT_PATH_PREFIX = re.compile(r"^/r/[^/]+/?")


async def _discover_extract_feeds_for_reddit(context: Context) -> tuple[Context, Result | None]:
    """New Reddit site has no links to RSS feeds => we construct them."""
    assert context.url is not None

    logger.info("discovering_reddit_extracting_feeds", url=context.url)

    f_url = construct_f_url(context.url)

    assert f_url is not None

    if f_url.host not in ("www.reddit.com", "reddit.com", "old.reddit.com"):
        # We are not interested in not reddit.com domains
        logger.info("discovering_reddit_not_reddit_domain")
        return context, None

    if f_url.host == "old.reddit.com":
        # Old Reddit site has marked RSS urls in the header
        logger.info("discovering_reddit_old_reddit_domain")
        return context, None

    match = _RE_REDDIT_PATH_PREFIX.match(str(f_url.path))

    if match is None:
        logger.info("discovering_reddit_not_reddit_path")
        return context, None

    base_path = match.group()

    if not base_path.endswith("/"):
        base_path += "/"

    f_url.path = f"{base_path}.rss"
    f_url.query = None

    # this check is required to stop recursion on _discover_check_candidate_links
    if str(f_url) == context.url:
        logger.info("discovering_reddit_same_url")
        return context, None

    logger.info("discovering_reddit_feed", feed_url=f_url)

    return context.replace(candidate_urls={str(f_url)}), None


# Note: we do not add internal feed discoverer here (like db check: url -> uid -> feed_id), because
#       - we do not expect significant performance improvement
#       - internal feed data (news list) may be slightly outdated (not containing the latest news)
_discoverers = [
    _discover_adjust_url,
    # This Reddit urls hack MUST go before loading url
    # because Reddit blocks access to the non-rss urls for bots
    #
    # TODO: Should we simulate browser behavior?
    #       Better not to, but it may be the only way to get the page data.
    _discover_extract_feeds_for_reddit,
    _discover_check_candidate_links,
    _discover_load_url,
    _discover_extract_feed_info,
    _discover_stop_recursion,
    _discover_create_soup,
    _discover_extract_feeds_from_links,
    _discover_check_candidate_links,
    _discover_extract_feeds_from_anchors,
    _discover_check_candidate_links,
    _discover_check_parent_urls,
]


async def discover(url: UnknownUrl | AbsoluteUrl, depth: int, discoverers: list[Discoverer] = _discoverers) -> Result:

    logger.info("discovering_start", url=url, depth=depth)

    context = Context(raw_url=UnknownUrl(url), depth=depth, discoverers=discoverers)

    for discoverer in discoverers:
        context, result = await discoverer(context)

        if result is not None:
            logger.info("discovering_finished", feeds_found=len(result.feeds))
            return result

    logger.info("discovering_finished", feeds_found=0)

    return Result(feeds=[], status=Status.no_feeds_found)



import uuid
from typing import Iterable

from ffun.core import logging
from ffun.core.postgresql import ExecuteType, execute
from ffun.domain.entities import EntryId, UserId
from ffun.markers.entities import Marker

logger = logging.get_module_logger()


async def set_marker(user_id: UserId, marker: Marker, entry_id: EntryId) -> None:
    sql = """
        INSERT INTO m_markers (id, user_id, marker, entry_id)
        VALUES (%(id)s, %(user_id)s, %(marker)s, %(entry_id)s)
        ON CONFLICT (user_id, marker, entry_id) DO NOTHING
        RETURNING id
    """

    results = await execute(sql, {"id": uuid.uuid4(), "user_id": user_id, "marker": marker, "entry_id": entry_id})

    if results:
        logger.business_event("marker_set", user_id=user_id, marker=marker, entry_id=entry_id)


async def remove_marker(user_id: UserId, marker: Marker, entry_id: EntryId) -> None:
    sql = """
        DELETE FROM m_markers
        WHERE user_id = %(user_id)s AND marker = %(marker)s AND entry_id = %(entry_id)s
        RETURNING id
    """

    resuts = await execute(sql, {"user_id": user_id, "marker": marker, "entry_id": entry_id})

    if resuts:
        logger.business_event("marker_removed", user_id=user_id, marker=marker, entry_id=entry_id)


# TODO: tests
async def get_markers(user_id: UserId, entries_ids: Iterable[EntryId]) -> dict[EntryId, set[Marker]]:
    sql = """
        SELECT entry_id, marker
        FROM m_markers
        WHERE user_id = %(user_id)s AND entry_id = ANY(%(entries_ids)s)
    """

    results = await execute(sql, {"user_id": user_id, "entries_ids": entries_ids})

    result: dict[EntryId, set[Marker]] = {}

    for row in results:
        entry_id = row["entry_id"]
        marker = Marker(row["marker"])

        if entry_id not in result:
            result[entry_id] = set()

        result[entry_id].add(marker)

    return result


# TODO: add business events?
async def tech_merge_markers(execute: ExecuteType, from_entry_id: EntryId, to_entry_id: EntryId) -> None:
    sql = """
        DELETE FROM m_markers AS m
        WHERE entry_id = %(from_entry_id)s
              AND EXISTS (SELECT 1 FROM m_markers WHERE entry_id = %(to_entry_id)s AND user_id = m.user_id)
    """

    await execute(sql, {"from_entry_id": from_entry_id, "to_entry_id": to_entry_id})

    sql = """
        UPDATE m_markers
        SET entry_id = %(to_entry_id)s
        WHERE entry_id = %(from_entry_id)s
    """

    await execute(sql, {"from_entry_id": from_entry_id, "to_entry_id": to_entry_id})


async def remove_markers_for_entries(entries_ids: Iterable[EntryId]) -> None:
    sql = """
        DELETE FROM m_markers
        WHERE entry_id = ANY(%(entries_ids)s)
    """

    await execute(sql, {"entries_ids": list(entries_ids)})



"""
markers-tables
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__: set[str] = set()

sql_create_markers_table = """
CREATE TABLE m_markers (
    id UUID PRIMARY KEY,
    user_id UUID NOT NULL,
    marker SMALLINT NOT NULL,
    entry_id UUID NOT NULL,
    created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW()
)
"""

sql_users_markers_index = """
CREATE UNIQUE INDEX idx_m_markers_user_id_marker_entry_id ON m_markers (user_id, entry_id, marker);
"""


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute(sql_create_markers_table)
    cursor.execute(sql_users_markers_index)


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("DROP TABLE m_markers")


steps = [step(apply_step, rollback_step)]



"""
index-to-search-by-value
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__ = {"20230701_01_7zBP0-settings-table"}


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("CREATE INDEX idx_us_settings_kind_value ON us_settings (kind, value)")


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("DROP INDEX idx_us_settings_kind_value")


steps = [step(apply_step, rollback_step)]



import pathlib
import re
import uuid

import frontmatter
import toml

from ffun.core import utils
from ffun.domain.entities import EntryId, SourceId
from ffun.library.entities import Entry
from ffun.processors_quality import errors
from ffun.processors_quality.entities import Attribution, ExpectedTags, ProcessorResult


class FrontmatterTOMLHandler(frontmatter.TOMLHandler):  # type: ignore
    FM_BOUNDARY = re.compile(r"^\-{3,}\s*$", re.MULTILINE)
    START_DELIMITER = END_DELIMITER = "---"


def id_to_name(id_: int) -> str:
    return f"{id_:04d}"


class KnowlegeBase:
    """Interface to operate with the base of test text.

    Catalogs structure:

    ./
    |-- news/
    |   |-- 0001.toml
    |   |-- 0002.toml
    |   |-- ...
    |-- tags-expected/
    |   |-- processor-1/
    |   |   |-- 0001.toml
    |   |   |-- 0002.toml
    |   |   |-- ...
    |-- tags-actual/
    |   |-- processor-1/
    |   |   |-- 0001.toml
    |   |   |-- 0002.toml
    |   |   |-- ...
    |-- tags-last/
    |   |-- processor-1/
    |   |   |-- 0001.toml
    |   |   |-- 0002.toml
    |   |   |-- ...
    """

    __slots__ = ("_dir_root", "_dir_news", "_dir_tags_expected", "_dir_tags_actual", "_dir_tags_last")

    def __init__(self, root: pathlib.Path) -> None:
        self._dir_root = root
        self._dir_news = root / "news"
        self._dir_tags_expected = root / "tags-expected"
        self._dir_tags_actual = root / "tags-actual"
        self._dir_tags_last = root / "tags-last"

    def get_news_entry(self, id_: int) -> Entry:
        entry_path = self._dir_news / f"{id_to_name(id_)}.toml"

        data, body = frontmatter.parse(entry_path.read_text(), handler=FrontmatterTOMLHandler())

        # attribution should be defined for the entry
        try:
            Attribution(**data["attribution"])
        except Exception as e:
            raise errors.AttributionNotDefined(message=str(e)) from e

        return Entry(
            id=EntryId(uuid.UUID(int=0)),
            source_id=SourceId(uuid.UUID(int=0)),
            title=data["title"],
            body=body,
            external_id="",
            external_url=data["external_url"],
            external_tags=data["external_tags"],
            cataloged_at=utils.now(),
            published_at=data["published_at"],
        )

    def get_expected_tags(self, processor: str, id_: int) -> ExpectedTags:
        tags_path = self._dir_tags_expected / processor / f"{id_to_name(id_)}.toml"

        data = toml.loads(tags_path.read_text())

        return ExpectedTags(must_have=set(data["tags_must_have"]), should_have=set(data["tags_should_have"]))

    def get_actual_results(self, processor: str, id_: int) -> ProcessorResult:
        tags_path = self._dir_tags_actual / processor / f"{id_to_name(id_)}.toml"
        return ProcessorResult(**toml.loads(tags_path.read_text()))

    def get_last_results(self, processor: str, id_: int) -> ProcessorResult:
        tags_path = self._dir_tags_last / processor / f"{id_to_name(id_)}.toml"
        return ProcessorResult(**toml.loads(tags_path.read_text()))

    def save_processor_result(self, processor: str, entry_id: int, result: ProcessorResult, actual: bool) -> None:
        (self._dir_tags_last / processor).mkdir(parents=True, exist_ok=True)

        last_path = self._dir_tags_last / processor / f"{id_to_name(entry_id)}.toml"

        content = toml.dumps(result.dict())

        last_path.write_text(content)

        if actual:
            (self._dir_tags_actual / processor).mkdir(parents=True, exist_ok=True)
            actual_path = self._dir_tags_actual / processor / f"{id_to_name(entry_id)}.toml"
            actual_path.write_text(content)

    def save_expected_data(
        self, processor: str, entry_id: int, tags_must_have: set[str], tags_should_have: set[str]
    ) -> None:
        if tags_must_have & tags_should_have:
            raise NotImplementedError("tags_must_have and tags_should_have should not intersect")

        (self._dir_tags_expected / processor).mkdir(parents=True, exist_ok=True)

        tags_path = self._dir_tags_expected / processor / f"{id_to_name(entry_id)}.toml"

        data = {"tags_must_have": list(sorted(tags_must_have)), "tags_should_have": list(sorted(tags_should_have))}

        tags_path.write_text(toml.dumps(data))

    def copy_last_to_actual(self, processor: str, entry_id: int) -> None:
        last_path = self._dir_tags_last / processor / f"{id_to_name(entry_id)}.toml"
        actual_path = self._dir_tags_actual / processor / f"{id_to_name(entry_id)}.toml"

        actual_path.write_text(last_path.read_text())

    def entry_ids(self) -> list[int]:
        return list(sorted(int(entry.stem) for entry in self._dir_news.glob("*.toml")))

    def validate_expected_tags(self) -> None:
        for entry_id in self.entry_ids():
            for processor in self._dir_tags_expected.iterdir():
                self.get_expected_tags(processor.name, entry_id)



"""
not-null-o_tags_properties-tag_id
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__ = {"20230617_03_de7vS-tags-properties"}


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()

    cursor.execute("ALTER TABLE o_tags_properties ALTER COLUMN tag_id SET NOT NULL")


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("ALTER TABLE o_tags_properties ALTER COLUMN tag_id DROP NOT NULL")


steps = [step(apply_step, rollback_step)]



import datetime
import uuid
from typing import Any, AsyncGenerator, Iterable

from pypika import PostgreSQLQuery

from ffun.core import logging
from ffun.core.postgresql import ExecuteType, execute, run_in_transaction
from ffun.domain.entities import EntryId, FeedId
from ffun.library.entities import Entry, FeedEntryLink
from ffun.library.settings import settings

logger = logging.get_module_logger()


def row_to_entry(row: dict[str, Any]) -> Entry:
    row["cataloged_at"] = row.pop("created_at")
    return Entry(**row)


@run_in_transaction
async def _catalog_entry(execute: ExecuteType, feed_id: FeedId, entry: Entry) -> None:
    sql_insert_entry = """
    INSERT INTO l_entries (id, source_id, title, body, external_id, external_url, external_tags, published_at)
    VALUES (%(id)s, %(source_id)s, %(title)s, %(body)s, %(external_id)s,
            %(external_url)s, %(external_tags)s, %(published_at)s)
    ON CONFLICT (source_id, external_id) DO NOTHING
    RETURNING id
    """

    sql_insert_feed_to_entry = """
    INSERT INTO l_feeds_to_entries (feed_id, entry_id)
    VALUES (%(feed_id)s, %(entry_id)s)
    ON CONFLICT (feed_id, entry_id) DO NOTHING
    """

    result = await execute(
        sql_insert_entry,
        {
            "id": entry.id,
            "source_id": entry.source_id,
            "title": entry.title,
            "body": entry.body,
            "external_id": entry.external_id,
            "external_url": entry.external_url,
            "external_tags": list(entry.external_tags),
            "published_at": entry.published_at,
        },
    )

    if result:
        entry_id = result[0]["id"]
    else:
        result = await execute(
            "SELECT id FROM l_entries WHERE source_id = %(source_id)s AND external_id = %(external_id)s",
            {"source_id": entry.source_id, "external_id": entry.external_id},
        )

        if result:
            entry_id = result[0]["id"]
        else:
            raise NotImplementedError("Can not find entry by source_id and external_id")

    await execute(sql_insert_feed_to_entry, {"feed_id": feed_id, "entry_id": entry_id})


async def catalog_entries(feed_id: FeedId, entries: Iterable[Entry]) -> None:
    for entry in entries:
        await _catalog_entry(feed_id, entry)


async def get_feed_links_for_entries(entries_ids: Iterable[EntryId]) -> dict[EntryId, list[FeedEntryLink]]:
    sql = """
    SELECT entry_id, feed_id, created_at
    FROM l_feeds_to_entries
    WHERE entry_id = ANY(%(entries_ids)s)
    """

    result = await execute(sql, {"entries_ids": list(entries_ids)})

    feeds_for_entries: dict[EntryId, list[FeedEntryLink]] = {}

    for row in result:
        entry_id = row["entry_id"]
        feed_id = row["feed_id"]
        created_at = row["created_at"]

        if entry_id not in feeds_for_entries:
            feeds_for_entries[entry_id] = []

        feeds_for_entries[entry_id].append(FeedEntryLink(feed_id=feed_id, entry_id=entry_id, created_at=created_at))

    return feeds_for_entries


async def find_stored_entries_for_feed(feed_id: FeedId, external_ids: Iterable[str]) -> set[str]:

    sql = """
    SELECT external_id
    FROM l_entries
    JOIN l_feeds_to_entries ON l_entries.id = l_feeds_to_entries.entry_id
    WHERE feed_id = %(feed_id)s AND external_id = ANY(%(external_ids)s)
    """

    rows = await execute(sql, {"external_ids": external_ids, "feed_id": feed_id})

    return {row["external_id"] for row in rows}


async def get_entries_by_ids(ids: Iterable[EntryId]) -> dict[EntryId, Entry | None]:

    sql_select_entries = """SELECT * FROM l_entries WHERE id = ANY(%(ids)s)"""

    rows = await execute(sql_select_entries, {"ids": ids})

    result: dict[EntryId, Entry | None] = {id: None for id in ids}

    for row in rows:
        result[row["id"]] = row_to_entry(row)

    return result


async def get_entries_by_filter(
    feeds_ids: Iterable[FeedId], limit: int, period: datetime.timedelta | None = None
) -> list[Entry]:
    if period is None:
        period = datetime.timedelta(days=100 * 365)

    # Here is an important logic implemented
    # When we are applying created time restriction
    # we are looking at a time of creating a link between entry and feed, not the real cataloged time
    # In most cases, those times should be nearly the same, but there is a possibility
    # that there will be aggregator-style feeds that include old entries as new.
    # In such cases, we'll show such an entry as new to a user.
    # We may want to change this logic in the future.

    sql = """
    SELECT le.*, re.max_created_at
    FROM (
        SELECT lfe.entry_id AS id, MAX(lfe.created_at) AS max_created_at
        FROM l_feeds_to_entries AS lfe
        WHERE lfe.created_at > NOW() - %(period)s
          AND lfe.feed_id = ANY(%(feeds_ids)s)
        GROUP BY lfe.entry_id
        ORDER BY MAX(lfe.created_at) DESC
        LIMIT %(limit)s
    ) AS re
    JOIN l_entries AS le ON le.id = re.id
    ORDER BY re.max_created_at DESC;
    """

    rows = await execute(sql, {"feeds_ids": feeds_ids, "period": period, "limit": limit})

    for row in rows:
        row.pop("max_created_at")

    return [row_to_entry(row) for row in rows]


async def get_entries_after_pointer(
    created_at: datetime.datetime, entry_id: EntryId, limit: int
) -> list[tuple[EntryId, datetime.datetime]]:
    sql = """
    SELECT id, created_at FROM l_entries
    WHERE created_at > %(created_at)s OR
          (created_at = %(created_at)s AND id > %(entry_id)s)
    ORDER BY created_at ASC, id ASC
    LIMIT %(limit)s
    """

    rows = await execute(sql, {"created_at": created_at, "entry_id": entry_id, "limit": limit})

    return [(row["id"], row["created_at"]) for row in rows]


# TODO: rewrite to use get_entries_after_pointer
async def all_entries_iterator(chunk: int) -> AsyncGenerator[Entry, None]:
    entry_id = uuid.UUID("00000000-0000-0000-0000-000000000000")

    sql = """
    SELECT * FROM l_entries
    WHERE %(entry_id)s < id
    ORDER BY id ASC
    LIMIT %(chunk)s
    """

    while True:
        rows = await execute(sql, {"entry_id": entry_id, "chunk": chunk})

        if not rows:
            break

        for row in rows:
            yield row_to_entry(row)

        entry_id = rows[-1]["id"]


async def update_external_url(entity_id: EntryId, url: str) -> None:
    sql = """
    UPDATE l_entries
    SET external_url = %(url)s
    WHERE id = %(entity_id)s
    """

    await execute(sql, {"entity_id": entity_id, "url": url})


async def unlink_feed_tail(feed_id: FeedId, offset: int | None = None) -> None:

    if offset is None:
        offset = settings.max_entries_per_feed

    # TODO: refactor to use https://www.psycopg.org/psycopg3/docs/api/cursors.html#psycopg.Cursor.rowcount
    sql = """
    WITH cte AS (
        SELECT feed_id, entry_id
        FROM l_feeds_to_entries
        WHERE feed_id = %(feed_id)s
        ORDER BY created_at DESC
        OFFSET %(offset)s
    )
    DELETE FROM l_feeds_to_entries
    USING cte
    WHERE l_feeds_to_entries.feed_id = cte.feed_id
      AND l_feeds_to_entries.entry_id = cte.entry_id
    RETURNING l_feeds_to_entries.entry_id AS entry_id
    """

    result = await execute(sql, {"feed_id": feed_id, "offset": offset})

    if not result:
        logger.info("feed_has_no_entries_tail", feed_id=feed_id, entries_limit=offset)
        return

    logger.info("feed_entries_tail_removed", feed_id=feed_id, entries_limit=offset, entries_removed=len(result))

    potential_orphanes = [row["entry_id"] for row in result]

    await try_mark_as_orphanes(potential_orphanes)


# TODO: metrics for orphaned entries and for feeds that produce them
async def try_mark_as_orphanes(entry_ids: Iterable[EntryId]) -> None:
    feed_links = await get_feed_links_for_entries(entry_ids)

    orphans = [entry_id for entry_id in entry_ids if entry_id not in feed_links]

    if not orphans:
        return

    query = PostgreSQLQuery.into("l_orphaned_entries").columns("entry_id")

    for entry_id in orphans:
        query = query.insert(entry_id)

    query = query.on_conflict("entry_id").do_nothing()

    await execute(str(query))


async def get_orphaned_entries(limit: int) -> set[EntryId]:
    sql = """
    SELECT entry_id
    FROM l_orphaned_entries
    LIMIT %(limit)s
    """

    rows = await execute(sql, {"limit": limit})

    return {row["entry_id"] for row in rows}


@run_in_transaction
async def remove_entries_by_ids(execute: ExecuteType, entry_ids: Iterable[EntryId]) -> None:
    ids = list(entry_ids)

    await execute("DELETE FROM l_feeds_to_entries WHERE entry_id = ANY(%(ids)s)", {"ids": ids})
    await execute("DELETE FROM l_entries WHERE id = ANY(%(ids)s)", {"ids": ids})
    await execute("DELETE FROM l_orphaned_entries WHERE entry_id = ANY(%(ids)s)", {"ids": ids})


async def count_total_entries() -> int:
    result = await execute("SELECT COUNT(*) FROM l_entries")
    return result[0]["count"]  # type: ignore



import pytest

from ffun.domain.domain import new_feed_id
from ffun.feeds import errors
from ffun.feeds.domain import get_feed


# get_feed function is checked in tests of other functions
class TestGetFeed:
    @pytest.mark.asyncio
    async def test_exception_if_no_feed_found(self) -> None:
        with pytest.raises(errors.NoFeedFound):
            await get_feed(new_feed_id())


# save_feeds function is checked in tests of other functions
class TestSaveFeeds:
    pass



from ffun.core.register import Entity, Register
from ffun.google.provider_interface import provider as google_provider
from ffun.llms_framework.entities import LLMProvider
from ffun.llms_framework.provider_interface import ProviderInterface, provider_test
from ffun.openai.provider_interface import provider as openai_provider


class Value(Entity):
    def __init__(self, key: str, provider: ProviderInterface) -> None:
        super().__init__(key=key)
        self.provider = provider


LLMProviders = Register[Value]

llm_providers: LLMProviders = Register()

llm_providers.add(Value(LLMProvider.test, provider_test))
llm_providers.add(Value(LLMProvider.openai, openai_provider))
llm_providers.add(Value(LLMProvider.google, google_provider))



import xml.etree.ElementTree as ET  # noqa
from typing import Generator

from ffun.domain.entities import UnknownUrl
from ffun.domain.urls import normalize_classic_unknown_url, to_feed_url, url_to_uid
from ffun.feeds.entities import Feed
from ffun.parsers.entities import FeedInfo


def _extract_body(data: str) -> ET.Element:
    root = ET.fromstring(data)  # noqa

    if root.tag != "opml":
        raise NotImplementedError(f"Unknown root tag: {root.tag}")

    body = root.find("body")

    if body is None:
        raise NotImplementedError("OPML file has no body tag")

    return body


def extract_feeds(data: str) -> list[FeedInfo]:
    feeds: list[FeedInfo] = []

    body = _extract_body(data)

    repeated_feeds = extract_feeds_records(body)

    for new_feed in repeated_feeds:
        for saved_feed in feeds:
            if new_feed.url == saved_feed.url:
                break
        else:
            feeds.append(new_feed)

    return feeds


def _extract_rss_feed(outline: ET.Element) -> FeedInfo | None:
    # TODO: here we may want to detect custom url scemes from the source of the OPML file
    #       for example, "newsletter:", and make something with them
    url = normalize_classic_unknown_url(UnknownUrl(outline.attrib["xmlUrl"]))

    if url is None:
        return None

    feed_url = to_feed_url(url)

    return FeedInfo(
        url=feed_url,
        title=outline.attrib.get("title", ""),
        description="",
        entries=[],
        uid=url_to_uid(feed_url),
    )


def extract_feeds_records(body: ET.Element) -> Generator[FeedInfo, None, None]:  # noqa: CCR001
    for outline in body:

        outline_type = outline.attrib.get("type")

        # small hack to fix OPML files without type attribute
        if outline_type is None and "xmlUrl" in outline.attrib:
            outline_type = "rss"

        if outline_type is not None:
            outline_type = outline_type.lower()

        if outline_type == "rss":
            if feed_info := _extract_rss_feed(outline):
                yield feed_info

            continue

        yield from extract_feeds_records(outline)


def create_opml(feeds: list[Feed]) -> str:

    feeds.sort(key=lambda feed: feed.title if feed.title is not None else "")

    opml = ET.Element("opml", version="2.0")

    head = ET.SubElement(opml, "head")
    title = ET.SubElement(head, "title")
    title.text = "Your subscriptions in feeds.fun"

    body = ET.SubElement(opml, "body")

    outline = ET.SubElement(body, "outline", {"title": "uncategorized", "text": "uncategorized"})

    for feed in feeds:
        feed_title = feed.title if feed.title is not None else "unknown"
        ET.SubElement(outline, "outline", {"title": feed_title, "text": feed_title, "type": "rss", "xmlUrl": feed.url})

    return ET.tostring(opml, encoding="utf-8", method="xml").decode("utf-8")  # type: ignore



import asyncio

from ffun.application.application import with_app
from ffun.cli.application import app
from ffun.core import logging
from ffun.library.domain import all_entries_iterator, normalize_entry

logger = logging.get_module_logger()


async def run(apply: bool, chunk: int, log_every_n: int) -> None:  # noqa: CCR001
    changes_counter = 0

    async with with_app():
        counter = 0

        async for entry in all_entries_iterator(chunk=chunk):
            counter += 1

            if counter % log_every_n == 0:
                logger.info("normalized_entries", number=counter)

            changes = await normalize_entry(entry, apply=apply)

            for change in changes:
                changes_counter += 1
                logger.info("change", number=changes_counter, change=change)

    logger.info("normalized_entries", number=counter)


@app.command()
def normalize_entries(apply: bool = False, chunk: int = 1000, log_every_n: int = 1000) -> None:
    asyncio.run(run(apply=apply, chunk=chunk, log_every_n=log_every_n))



import uuid

import pytest
from pytest_mock import MockerFixture
from structlog.testing import capture_logs

from ffun.core.tests.helpers import assert_logs, assert_logs_has_business_event
from ffun.domain.entities import UserId
from ffun.domain.urls import url_to_uid
from ffun.feeds import domain as f_domain
from ffun.feeds import entities as f_entities
from ffun.feeds_collections.collections import collections
from ffun.feeds_collections.entities import CollectionId
from ffun.feeds_links import domain as fl_domain
from ffun.library import domain as l_domain
from ffun.library import entities as l_entities
from ffun.loader.domain import check_proxies_availability, detect_orphaned, process_feed, store_entries, sync_feed_info
from ffun.loader.settings import Proxy, settings
from ffun.parsers import entities as p_entities
from ffun.parsers.tests import make as p_make


def assert_entriy_equal_to_info(entry_info: p_entities.EntryInfo, entry: l_entities.Entry) -> None:
    assert entry.title == entry_info.title
    assert entry.body == entry_info.body
    assert entry.external_id == entry_info.external_id
    assert entry.external_url == entry_info.external_url
    assert entry.external_tags == entry_info.external_tags
    assert entry.published_at == entry_info.published_at


class TestDetectOrphaned:
    @pytest.mark.asyncio
    async def test_is_orphaned(self, saved_feed_id: f_entities.FeedId) -> None:
        assert await detect_orphaned(saved_feed_id)

        loaded_feed = await f_domain.get_feed(saved_feed_id)

        assert loaded_feed.state == f_entities.FeedState.orphaned

    @pytest.mark.asyncio
    async def test_is_orphaned_but_from_collection(
        self,
        saved_feed_id: f_entities.FeedId,
        collection_id_for_test_feeds: CollectionId,
    ) -> None:
        await collections.add_test_feed_to_collections(collection_id_for_test_feeds, saved_feed_id)

        assert not await detect_orphaned(saved_feed_id)

        loaded_feed = await f_domain.get_feed(saved_feed_id)

        assert loaded_feed.state != f_entities.FeedState.orphaned

    @pytest.mark.asyncio
    async def test_not_orphaned(self, internal_user_id: UserId, saved_feed_id: f_entities.FeedId) -> None:
        await fl_domain.add_link(internal_user_id, saved_feed_id)

        assert not await detect_orphaned(saved_feed_id)

        loaded_feed = await f_domain.get_feed(saved_feed_id)

        assert loaded_feed.state != f_entities.FeedState.orphaned


class TestSyncFeedInfo:
    @pytest.mark.asyncio
    async def test_no_sync_required(self, saved_feed: f_entities.Feed, mocker: MockerFixture) -> None:
        update_feed_info = mocker.patch("ffun.feeds.domain.update_feed_info")

        assert saved_feed.title
        assert saved_feed.description

        feed_info = p_entities.FeedInfo(
            url=saved_feed.url,
            title=saved_feed.title,
            description=saved_feed.description,
            entries=[],
            uid=url_to_uid(saved_feed.url),
        )

        await sync_feed_info(saved_feed, feed_info)

        update_feed_info.assert_not_called()

    @pytest.mark.asyncio
    async def test_sync_required(self, saved_feed: f_entities.Feed) -> None:
        feed_info = p_entities.FeedInfo(
            url=saved_feed.url,
            title=uuid.uuid4().hex,
            description=uuid.uuid4().hex,
            entries=[],
            uid=url_to_uid(saved_feed.url),
        )

        await sync_feed_info(saved_feed, feed_info)

        loaded_feed = await f_domain.get_feed(saved_feed.id)

        assert loaded_feed.title == feed_info.title
        assert loaded_feed.description == feed_info.description

    @pytest.mark.asyncio
    async def test_sync_required__collections(
        self, collection_id_for_test_feeds: CollectionId, saved_feed: f_entities.Feed
    ) -> None:
        await collections.add_test_feed_to_collections(collection_id_for_test_feeds, saved_feed.id)

        feed_info = p_entities.FeedInfo(
            url=saved_feed.url,
            title=uuid.uuid4().hex,
            description=uuid.uuid4().hex,
            entries=[],
            uid=url_to_uid(saved_feed.url),
        )

        await sync_feed_info(saved_feed, feed_info)

        loaded_feed = await f_domain.get_feed(saved_feed.id)

        assert loaded_feed.title != feed_info.title
        assert loaded_feed.description != feed_info.description

        collection_feed_info = collections.get_feed_info(saved_feed.id)

        assert loaded_feed.title == collection_feed_info.title
        assert loaded_feed.description == collection_feed_info.description


class TestStoreEntries:
    @pytest.mark.asyncio
    async def test_no_entries(self, saved_feed: f_entities.Feed) -> None:
        with capture_logs() as logs:
            await store_entries(saved_feed, [])

        assert_logs(logs, news_entries_stored=0)

        entries = await l_domain.get_entries_by_filter([saved_feed.id], limit=1)

        assert not entries

    @pytest.mark.asyncio
    async def test_save_new_entries(self, saved_feed: f_entities.Feed) -> None:
        n = 3

        entry_infos = [p_make.fake_entry_info() for _ in range(n)]

        with capture_logs() as logs:
            await store_entries(saved_feed, entry_infos)

        assert_logs_has_business_event(
            logs, "news_entries_stored", user_id=None, feed_id=str(saved_feed.id), entries_number=n
        )

        loaded_entries = await l_domain.get_entries_by_filter([saved_feed.id], limit=n + 1)

        assert len(loaded_entries) == 3

        entry_infos.sort(key=lambda e: e.title)
        loaded_entries.sort(key=lambda e: e.title)

        for entry_info, entry in zip(entry_infos, loaded_entries):
            assert entry.source_id == saved_feed.source_id
            assert_entriy_equal_to_info(entry_info, entry)

    @pytest.mark.asyncio
    async def test_save_in_parts(self, saved_feed: f_entities.Feed) -> None:
        n = 5
        m = 3

        assert m < n

        entry_infos = [p_make.fake_entry_info() for _ in range(n)]
        entry_infos.sort(key=lambda e: e.title)

        with capture_logs() as logs:
            await store_entries(saved_feed, entry_infos[:m])

        assert_logs_has_business_event(
            logs, "news_entries_stored", user_id=None, feed_id=str(saved_feed.id), entries_number=m
        )

        loaded_entries = await l_domain.get_entries_by_filter([saved_feed.id], limit=n + 1)

        assert len(loaded_entries) == m

        loaded_entries.sort(key=lambda e: e.title)

        for entry_info, entry in zip(entry_infos, loaded_entries):
            assert entry.source_id == saved_feed.source_id
            assert_entriy_equal_to_info(entry_info, entry)

        with capture_logs() as logs:
            await store_entries(saved_feed, entry_infos)

        assert_logs_has_business_event(
            logs, "news_entries_stored", user_id=None, feed_id=str(saved_feed.id), entries_number=n - m
        )

        loaded_entries = await l_domain.get_entries_by_filter([saved_feed.id], limit=n + 1)

        assert len(loaded_entries) == n

        loaded_entries.sort(key=lambda e: e.title)

        for entry_info, entry in zip(entry_infos, loaded_entries):
            assert entry.source_id == saved_feed.source_id
            assert_entriy_equal_to_info(entry_info, entry)


class TestProcessFeed:
    @pytest.mark.asyncio
    async def test_orphaned_feed(self, saved_feed: f_entities.Feed) -> None:
        assert await detect_orphaned(saved_feed.id)

        await process_feed(feed=saved_feed)

        loaded_feed = await f_domain.get_feed(saved_feed.id)

        assert loaded_feed.state == f_entities.FeedState.orphaned

        loaded_entries = await l_domain.get_entries_by_filter([saved_feed.id], limit=1)

        assert not loaded_entries

    @pytest.mark.asyncio
    async def test_can_not_extract_feed(
        self, internal_user_id: UserId, saved_feed: f_entities.Feed, mocker: MockerFixture
    ) -> None:
        extract_feed_info = mocker.patch("ffun.loader.domain.extract_feed_info", return_value=None)

        await fl_domain.add_link(internal_user_id, saved_feed.id)

        await process_feed(feed=saved_feed)

        extract_feed_info.assert_called_once_with(feed_id=saved_feed.id, feed_url=saved_feed.url)

        loaded_entries = await l_domain.get_entries_by_filter([saved_feed.id], limit=1)

        assert not loaded_entries

    @pytest.mark.asyncio
    async def test_success(self, internal_user_id: UserId, saved_feed: f_entities.Feed, mocker: MockerFixture) -> None:
        n = 3

        entry_infos = [p_make.fake_entry_info() for _ in range(n)]
        entry_infos.sort(key=lambda e: e.title)

        assert saved_feed.title
        assert saved_feed.description

        feed_info = p_entities.FeedInfo(
            url=saved_feed.url,
            title=saved_feed.title,
            description=saved_feed.description,
            entries=entry_infos,
            uid=url_to_uid(saved_feed.url),
        )

        extract_feed_info = mocker.patch("ffun.loader.domain.extract_feed_info", return_value=feed_info)

        await fl_domain.add_link(internal_user_id, saved_feed.id)

        with capture_logs() as logs:
            await process_feed(feed=saved_feed)

        assert_logs(logs, feed_has_no_entries_tail=1, feed_entries_tail_removed=0)

        extract_feed_info.assert_called_once_with(feed_id=saved_feed.id, feed_url=saved_feed.url)

        loaded_entries = await l_domain.get_entries_by_filter([saved_feed.id], limit=n + 1)

        assert len(loaded_entries) == n

        entry_infos.sort(key=lambda e: e.title)
        loaded_entries.sort(key=lambda e: e.title)

        for entry_info, entry in zip(entry_infos, loaded_entries):
            assert entry.source_id == saved_feed.source_id
            assert_entriy_equal_to_info(entry_info, entry)

    @pytest.mark.asyncio
    async def test_remove_too_long_entries_tail__when_feed_is_updated(
        self, internal_user_id: UserId, saved_feed: f_entities.Feed, mocker: MockerFixture
    ) -> None:
        n = 5
        m = 3

        entry_infos = [p_make.fake_entry_info() for _ in range(n)]
        entry_infos.sort(key=lambda e: e.title)

        assert saved_feed.title
        assert saved_feed.description

        feed_info = p_entities.FeedInfo(
            url=saved_feed.url,
            title=saved_feed.title,
            description=saved_feed.description,
            entries=entry_infos,
            uid=url_to_uid(saved_feed.url),
        )

        mocker.patch("ffun.loader.domain.extract_feed_info", return_value=feed_info)
        mocker.patch("ffun.library.settings.settings.max_entries_per_feed", m)

        await fl_domain.add_link(internal_user_id, saved_feed.id)

        with capture_logs() as logs:
            await process_feed(feed=saved_feed)

        assert_logs(logs, feed_has_no_entries_tail=0, feed_entries_tail_removed=1)

        loaded_entries = await l_domain.get_entries_by_filter([saved_feed.id], limit=n + 1)

        assert len(loaded_entries) == m

    @pytest.mark.asyncio
    async def test_remove_too_long_entries_tail__when_feed_is_not_updated(
        self, internal_user_id: UserId, saved_feed: f_entities.Feed, mocker: MockerFixture
    ) -> None:
        n = 5
        m = 3

        entry_infos = [p_make.fake_entry_info() for _ in range(n)]
        entry_infos.sort(key=lambda e: e.title)

        assert saved_feed.title
        assert saved_feed.description

        await store_entries(saved_feed, entry_infos)

        mocker.patch("ffun.loader.domain.extract_feed_info", return_value=None)
        mocker.patch("ffun.library.settings.settings.max_entries_per_feed", m)

        await fl_domain.add_link(internal_user_id, saved_feed.id)

        with capture_logs() as logs:
            await process_feed(feed=saved_feed)

        assert_logs(logs, feed_has_no_entries_tail=0, feed_entries_tail_removed=1)

        loaded_entries = await l_domain.get_entries_by_filter([saved_feed.id], limit=n + 1)

        assert len(loaded_entries) == m


class TestCheckProxiesAvailability:
    @pytest.mark.asyncio
    async def test_no_proxies(self, mocker: MockerFixture) -> None:
        mocker.patch("ffun.loader.settings.settings.proxies", [])

        is_proxy_available = mocker.patch("ffun.loader.operations.is_proxy_available")

        await check_proxies_availability()

        is_proxy_available.assert_not_called()

    @pytest.mark.asyncio
    async def test_proxies(self, mocker: MockerFixture) -> None:
        n = 3

        proxies = [Proxy(name=uuid.uuid4().hex, url=None) for _ in range(n)]

        mocker.patch("ffun.loader.settings.settings.proxies", proxies)
        mocker.patch("ffun.loader.domain._user_agent", "test user agent")

        is_proxy_available = mocker.patch("ffun.loader.operations.is_proxy_available")

        await check_proxies_availability()

        for proxy in proxies:
            is_proxy_available.assert_any_call(
                proxy=proxy, anchors=settings.proxy_anchors, user_agent="test user agent"
            )



from collections import Counter
from typing import Iterable

from bidict import bidict

from ffun.core.postgresql import ExecuteType, execute, run_in_transaction, transaction
from ffun.domain.entities import EntryId, TagId, TagUid
from ffun.ontology import operations
from ffun.ontology.entities import ProcessorTag, Tag, TagCategory, TagPropertyType
from ffun.tags import converters

count_total_tags = operations.count_total_tags
count_total_tags_per_type = operations.count_total_tags_per_type
count_total_tags_per_category = operations.count_total_tags_per_category


_tags_cache: bidict[TagUid, TagId] = bidict()


async def get_id_by_uid(tag: TagUid) -> TagId:
    if tag in _tags_cache:
        return _tags_cache[tag]

    tag_id = await operations.get_or_create_id_by_tag(tag)

    _tags_cache[tag] = tag_id

    return tag_id


async def get_ids_by_uids(tags: Iterable[TagUid]) -> dict[TagUid, TagId]:
    return {tag: await get_id_by_uid(tag) for tag in tags}


async def get_tags_by_ids(ids: Iterable[TagId]) -> dict[TagId, TagUid]:
    result = {}

    tags_to_request = []

    for tag_id in ids:
        if tag_id in _tags_cache.inverse:
            result[tag_id] = _tags_cache.inverse[tag_id]
        else:
            tags_to_request.append(tag_id)

    if not tags_to_request:
        return result

    missed_tags = await operations.get_tags_by_ids(tags_to_request)

    _tags_cache.inverse.update(missed_tags)

    result.update(missed_tags)

    return result


# TODO: tests
async def normalize_tags(tags: Iterable[ProcessorTag]) -> dict[str, TagUid]:
    return {tag.raw_uid: converters.normalize(tag.raw_uid) for tag in tags}


# TODO: in the future we could split this function into two separate functions
#       1. tags & properties normalization
#       2. saving tags & properties to the database
async def apply_tags_to_entry(entry_id: EntryId, processor_id: int, tags: Iterable[ProcessorTag]) -> None:
    """Apply tags to entry.

    Function expects raw tags from processors => after normalization we could have duplicates.
    => Function skips duplicated tags.
    """
    raw_to_uids = await normalize_tags(tags)

    uids_to_ids = await get_ids_by_uids(raw_to_uids.values())

    properties = []

    processed_tags = set()

    for tag in tags:
        tag_id = uids_to_ids[raw_to_uids[tag.raw_uid]]

        if tag_id in processed_tags:
            # skip duplicated tags
            continue

        processed_tags.add(tag_id)

        properties.extend(tag.build_properties_for(tag_id, processor_id=processor_id))

    async with transaction() as execute:
        await operations.apply_tags(execute, entry_id, processor_id, uids_to_ids.values())
        await operations.apply_tags_properties(execute, properties)


async def get_tags_ids_for_entries(entries_ids: list[EntryId]) -> dict[EntryId, set[TagId]]:
    return await operations.get_tags_for_entries(execute, entries_ids)


async def get_tags_info(tags_ids: Iterable[TagId]) -> dict[TagId, Tag]:  # noqa: CCR001
    # we expect that properties will be sorted by date from the newest to the oldest
    properties = await operations.get_tags_properties(tags_ids)

    tags_by_ids = await get_tags_by_ids(tags_ids)

    info = {tag_id: Tag(id=tag_id, name=converters.verbose(tags_by_ids[tag_id])) for tag_id in tags_ids}

    # TODO: implement more complex merging
    for property in properties:
        tag = info[property.tag_id]

        if property.type == TagPropertyType.link:
            if tag.link is None:
                tag.link = property.value
            continue

        if property.type == TagPropertyType.categories:
            tag.categories.update(TagCategory(cat) for cat in property.value.split(","))
            continue

        raise NotImplementedError(f"Unknown property type: {property.type}")

    return info


@run_in_transaction
async def remove_relations_for_entries(execute: ExecuteType, entries_ids: list[EntryId]) -> None:
    await operations.remove_relations_for_entries(execute, entries_ids)


@run_in_transaction
async def tech_copy_relations(execute: ExecuteType, entry_from_id: EntryId, entry_to_id: EntryId) -> None:
    await operations.tech_copy_relations(execute, entry_from_id, entry_to_id)


def _inplace_filter_out_entry_tags(
    entry_tag_ids: dict[EntryId, set[TagId]], must_have_tags: set[TagId], min_tag_count: int
) -> set[TagId]:
    """Function modifies `entry_tag_ids` in place and returns a final set of tags that still there."""
    tags_count: Counter[TagId] = Counter()
    for tags in entry_tag_ids.values():
        tags_count.update(tags)

    tags_to_exclude = {tag_id for tag_id, count in tags_count.items() if count < min_tag_count}
    tags_to_exclude -= must_have_tags

    whole_tags = set(tags_count.keys()) - tags_to_exclude

    for entry_tags in entry_tag_ids.values():
        entry_tags.intersection_update(whole_tags)

    return whole_tags


# TODO: tests
async def prepare_tags_for_entries(
    entry_ids: list[EntryId], must_have_tags: set[TagId], min_tag_count: int
) -> tuple[dict[EntryId, set[TagId]], dict[TagId, TagUid]]:

    entry_tag_ids = await get_tags_ids_for_entries(entry_ids)

    whole_tags = _inplace_filter_out_entry_tags(entry_tag_ids, must_have_tags, min_tag_count)

    tag_mapping = await get_tags_by_ids(whole_tags)

    return entry_tag_ids, tag_mapping



import pytest

from ffun.librarian.tag_extractors import dog_tags_extractor


class TestDogTagsExtractor:

    raw_answer = """\
Here is a list of tags to describe the provided text, categorized as requested:

### Topics
- @singlwordtag
- @multi-word-tag
- @duplicated-tag


### Areas
- @tag-with-34-numbers-324
- @duplicated-tag

@tag-in-the-line-1, @tag-in-the-line-2, @tag-in-the-line-3

@__broken_taG-that-should-be-found

@-_this-----tag-will-be-found-too

this-tag-will-not-be-found
andthistagtoo

и-этот-тег-тоже-не-найдут

-and-this-one-tag-is-wrong
"""

    @pytest.mark.asyncio
    async def test(self) -> None:
        tags = dog_tags_extractor(self.raw_answer)

        expected_tags = {
            "singlwordtag",
            "multi-word-tag",
            "duplicated-tag",
            "tag-with-34-numbers-324",
            "tag-in-the-line-1",
            "tag-in-the-line-2",
            "tag-in-the-line-3",
            "__broken_tag-that-should-be-found",
            "-_this-----tag-will-be-found-too",
        }

        assert tags == expected_tags



import enum
from decimal import Decimal

from ffun.user_settings import types
from ffun.user_settings.values import Value, user_settings


class UserSetting(enum.IntEnum):
    openai_api_key = 1

    # openai_max_tokens_in_month = 2  # noqa

    hide_message_about_setting_up_key = 3

    process_entries_not_older_than = 4

    # openai_allow_use_key_for_collections = 5  # noqa

    max_tokens_cost_in_month = 6

    gemini_api_key = 7
    test_api_key = 8

    hide_message_about_adding_collections = 9
    hide_message_check_your_feed_urls = 10


user_settings.add(
    Value(
        key=UserSetting.openai_api_key,
        name="OpenAI API key",
        type=types.Secret(),
        default="",
    )
)


user_settings.add(
    Value(
        key=UserSetting.gemini_api_key,
        name="Gemini API key",
        type=types.Secret(),
        default="",
    )
)


user_settings.add(
    Value(
        key=UserSetting.test_api_key,
        name="Test API key, YOU SHOULD NOT SEE THIS, tell to developers",
        type=types.Secret(),
        default="",
    )
)


user_settings.add(
    Value(
        key=UserSetting.max_tokens_cost_in_month,
        name="Max spending (USD/month)",
        type=types.Decimal(),
        default=Decimal("10"),
    )
)


user_settings.add(
    Value(
        key=UserSetting.hide_message_about_setting_up_key,
        name="Hide message about setting up API keys",
        type=types.Boolean(),
        default=False,
    )
)


user_settings.add(
    Value(
        key=UserSetting.process_entries_not_older_than,
        name="Process entries not older than N days",
        type=types.Integer(),
        default=1,
    )
)


user_settings.add(
    Value(
        key=UserSetting.hide_message_about_adding_collections,
        name="Hide message about adding feeds from collections",
        type=types.Boolean(),
        default=False,
    )
)


user_settings.add(
    Value(
        key=UserSetting.hide_message_check_your_feed_urls,
        name="Hide message about checking your feed URLs",
        type=types.Boolean(),
        default=False,
    )
)



import uuid

from ffun.domain.entities import UserId
from ffun.users import operations
from ffun.users.entities import Service


async def fake_user_id() -> UserId:
    external_id = f"fake-user#{uuid.uuid4()}"
    return await operations.add_mapping(Service.supertokens, external_id)


async def n_users(n: int) -> list[UserId]:
    return [await fake_user_id() for _ in range(n)]



"""
optimize_l_entries
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__ = {"20230514_01_Bwb35-processed-state"}


sql_remove_column = """
ALTER TABLE l_entries DROP COLUMN cataloged_at;
"""

sql_return_column = """
ALTER TABLE l_entries ADD COLUMN cataloged_at TIMESTAMP WITH TIME ZONE DEFAULT NULL;
"""

sql_return_column_values = """
UPDATE l_entries
SET cataloged_at = created_at;
"""

sql_return_not_null = """
ALTER TABLE l_entries ALTER COLUMN cataloged_at SET NOT NULL;
"""

sql_create_index = """
CREATE INDEX idx_l_entries_created_at_feed_id ON l_entries (created_at, feed_id);
"""


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()

    cursor.execute(sql_remove_column)
    cursor.execute(sql_create_index)


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()

    cursor.execute("DROP INDEX idx_l_entries_created_at_feed_id")

    cursor.execute(sql_return_column)
    cursor.execute(sql_return_column_values)
    cursor.execute(sql_return_not_null)


steps = [step(apply_step, rollback_step)]
